import os
import re

import pytest

from upsies import errors
from upsies.utils import predbs
from upsies.utils.types import SceneCheckResult


@pytest.fixture
def multipredb():
    return predbs.MultiPredbApi()


@pytest.mark.parametrize(
    argnames='release_name, exp_return_value',
    argvalues=(
        pytest.param(
            'Dellamorte.Dellamore.1994.1080p.BluRay.x264.mkv',
            SceneCheckResult.false,
            id='Movie: NOGROUP',
        ),
        pytest.param(
            'Dellamorte.Dellamore.1994.1080p.BluRay.x264-NOGROUP.mkv',
            SceneCheckResult.false,
            id='Movie: NOGROUP',
        ),
        pytest.param(
            'Dellamorte.Dellamore.1994.1080p.BluRay.x264-NoGrp.mkv',
            SceneCheckResult.false,
            id='Movie: NOGROUP',
        ),

        pytest.param(
            'Dellamorte.Dellamore.1994.1080p.BluRay.x264-LiViDiTY.mkv',
            SceneCheckResult.true,
            id='Movie: Properly named file (file name is release name with file extension)',
        ),
        pytest.param(
            'Dellamorte.Dellamore.1994.1080p.Blu-ray.x264-LiViDiTY.mkv',
            SceneCheckResult.true,
            id='Movie: Properly named file (renamed)',
        ),
        pytest.param(
            'dellamorte.dellamore.1994.1080p.blu-ray.x264-lividity.mkv',
            SceneCheckResult.true,
            id='Movie: Properly named file (renamed)',
        ),

        pytest.param(
            'Dellamorte.Dellamore.1994.1080p.BluRay.x264-LiViDiTY/Dellamorte.Dellamore.1994.1080p.BluRay.x264-LiViDiTY.mkv',
            SceneCheckResult.true,
            id='Movie: Properly named file in properly named directory (not renamed)',
        ),
        pytest.param(
            'Dellamorte.Dellamore.1994.1080p.Blu-ray.x264-LiViDiTY/Dellamorte.Dellamore.1994.1080p.BluRay.x264-LiViDiTY.mkv',
            SceneCheckResult.true,
            id='Movie: Properly named file in properly named directory (renamed directory)',
        ),
        pytest.param(
            'Dellamorte.Dellamore.1994.1080p.BluRay.x264-LiViDiTY/Dellamorte.Dellamore.1994.1080p.Blu-ray.x264-LiViDiTY.mkv',
            SceneCheckResult.true,
            id='Movie: Properly named file in properly named directory (renamed file)',
        ),

        pytest.param(
            'path/to/ly-dellmdm1080p.mkv',
            SceneCheckResult.false,
            id='Movie: Abbreviated file without usably named parent directory',
        ),

        pytest.param(
            'Dellamorte.Dellamore.1994.1080p.BluRay.x264-LiViDiTY/ly-dellmdm1080p.mkv',
            SceneCheckResult.true,
            id='Movie: Abbreviated file in usably named directory',
        ),
        pytest.param(
            'Dellamorte.Dellamore.1994.1080p.Blu-ray.x264-LiViDiTY/ly-dellmdm1080p.mkv',
            SceneCheckResult.true,
            id='Movie: Abbreviated file in usably named directory',
        ),
        pytest.param(
            'dellamorte.dellamore.1994.1080p.blu-ray.x264-lividity/ly-dellmdm1080p.mkv',
            SceneCheckResult.true,
            id='Movie: Abbreviated file in usably named directory',
        ),

        pytest.param(
            'Bored.to.Death.S01.EXTRAS.720p.BluRay.x264-iNGOT',
            SceneCheckResult.true,
            id='Series: Scene released season pack',
        ),
        pytest.param(
            'bored.to.death.s01.720p.bluray.x264-ingot',
            SceneCheckResult.true,
            id='Series: Scene released season pack',
        ),
        pytest.param(
            'Bored.to.Death.S01.Jonathan.Amess.Brooklyn.720p.BluRay.x264-iNGOT.mkv',
            SceneCheckResult.true,
            id='Series: Scene released season pack (Extras)',
        ),
        pytest.param(
            'bored.to.death.s01.720p.bluray.x264-ingot.mkv',
            SceneCheckResult.true,
            id='Series: Scene released season pack (Extras)',
        ),

        pytest.param(
            'Justified.S04E01.720p.BluRay.x264-REWARD',
            SceneCheckResult.true,
            id='Series: Scene released single episodes, existing release',
        ),

        pytest.param(
            'Justified.S04E99.720p.BluRay.x264-REWARD',
            SceneCheckResult.false,
            id='Series: Scene released single episodes, non-existing release',
        ),

        pytest.param(
            'Justified.S04.720p.BluRay.x264-REWARD',
            SceneCheckResult.true,
            id='Series: Scene released single episodes, season pack',
        ),

        pytest.param(
            'Justified.S02.720p.BluRay.x264-REWARD',
            SceneCheckResult.false,
            id='Series: Missing season from scene group',
        ),

        pytest.param(
            'Justified.720p.BluRay.x264-REWARD',
            SceneCheckResult.true,
            id='Series: Provided release name is missing: episodes, source',
        ),

        pytest.param(
            'Crank.2.High.Voltage.720p.BluRay.x264-iNFAMOUS',
            SceneCheckResult.true,
            id='Movie: Original scene release name is missing information: year',
        ),

        pytest.param(
            'Friends.S10.1080p.BluRay.x264-TENEIGHTY',
            SceneCheckResult.true,
            id='Series: Lots of seasons and episodes',
        ),

        pytest.param(
            'The.Fall.Guy.S02.480p.DVDRip.XviD.AAC-nodlabs',
            SceneCheckResult.true,
            id='Resolution is ignored for DVDRip',
        ),

        pytest.param(
            'Foo.2011.1080p.Bluray.DD5.1.x264-NOTASCENEGROUP.mkv',
            SceneCheckResult.false,
            id='Non-scene movie release',
        ),

        pytest.param(
            'Damnation.S01.720p.AMZN.WEB-DL.DDP5.1.H.264-AJP69',
            SceneCheckResult.false,
            id='Non-scene season release',
        ),

        pytest.param(
            'Damnation.S01E03.One.Penny.720p.AMZN.WEB-DL.DD+5.1.H.264-AJP69.mkv',
            SceneCheckResult.false,
            id='Non-scene episode release',
        ),
    ),
)
@pytest.mark.asyncio
async def test_is_scene_release(release_name, exp_return_value, multipredb, store_response):
    assert await multipredb.is_scene_release(release_name) is exp_return_value


@pytest.mark.parametrize(
    argnames='release_name, exp_return_value',
    argvalues=(
        pytest.param(
            'Dellamorte.Dellamore.1994.1080p.BluRay.x264-LiViDiTY',
            {'ly-dellmdm1080p.mkv': {'release_name': 'Dellamorte.Dellamore.1994.1080p.BluRay.x264-LiViDiTY',
                                     'file_name': 'ly-dellmdm1080p.mkv',
                                     'size': 7044061767,
                                     'crc': 'B59CE234'}},
            id='Looking for single-file release when original release was single-file',
        ),

        pytest.param(
            'Bored.to.Death.S01.EXTRAS.720p.BluRay.x264-iNGOT',
            {'Bored.to.Death.S01.Jonathan.Amess.Brooklyn.720p.BluRay.x264-iNGOT.mkv':
             {'release_name': 'Bored.to.Death.S01.EXTRAS.720p.BluRay.x264-iNGOT',
              'file_name': 'Bored.to.Death.S01.Jonathan.Amess.Brooklyn.720p.BluRay.x264-iNGOT.mkv',
              'size': 518652629,
              'crc': '497277ED'},
             'Bored.to.Death.S01.Making.of.720p.BluRay.x264-iNGOT.mkv':
             {'release_name': 'Bored.to.Death.S01.EXTRAS.720p.BluRay.x264-iNGOT',
              'file_name': 'Bored.to.Death.S01.Making.of.720p.BluRay.x264-iNGOT.mkv',
              'size': 779447228,
              'crc': '8C8C0AE7'},
             'Bored.to.Death.S01E03.Deleted.Scene.720p.BluRay.x264-iNGOT.mkv':
             {'release_name': 'Bored.to.Death.S01.EXTRAS.720p.BluRay.x264-iNGOT',
              'file_name': 'Bored.to.Death.S01E03.Deleted.Scene.720p.BluRay.x264-iNGOT.mkv',
              'size': 30779540,
              'crc': '0ECBEBE8'},
             'Bored.to.Death.S01E04.Deleted.Scene.720p.BluRay.x264-iNGOT.mkv':
             {'release_name': 'Bored.to.Death.S01.EXTRAS.720p.BluRay.x264-iNGOT',
              'file_name': 'Bored.to.Death.S01E04.Deleted.Scene.720p.BluRay.x264-iNGOT.mkv',
              'size': 138052914,
              'crc': 'E4E41C29'},
             'Bored.to.Death.S01E08.Deleted.Scene.1.720p.BluRay.x264-iNGOT.mkv':
             {'release_name': 'Bored.to.Death.S01.EXTRAS.720p.BluRay.x264-iNGOT',
              'file_name': 'Bored.to.Death.S01E08.Deleted.Scene.1.720p.BluRay.x264-iNGOT.mkv',
              'size': 68498554,
              'crc': '085C6946'},
             'Bored.to.Death.S01E08.Deleted.Scene.2.720p.BluRay.x264-iNGOT.mkv':
             {'release_name': 'Bored.to.Death.S01.EXTRAS.720p.BluRay.x264-iNGOT',
              'file_name': 'Bored.to.Death.S01E08.Deleted.Scene.2.720p.BluRay.x264-iNGOT.mkv',
              'size': 45583011,
              'crc': '7B822E59'}},
            id='Looking for multi-file release when original release was multi-file',
        ),

        pytest.param(
            'Fawlty.Towers.S01.720p.BluRay.x264-SHORTBREHD',
            {'Fawlty.Towers.S01E01.720p.BluRay.x264-SHORTBREHD.mkv':
             {'release_name': 'Fawlty.Towers.S01E01.720p.BluRay.x264-SHORTBREHD',
              'file_name': 'Fawlty.Towers.S01E01.720p.BluRay.x264-SHORTBREHD.mkv',
              'size': 1559430969, 'crc': 'B334CBEA'},
             'Fawlty.Towers.S01E02.720p.BluRay.x264-SHORTBREHD.mkv':
             {'release_name': 'Fawlty.Towers.S01E02.720p.BluRay.x264-SHORTBREHD',
              'file_name': 'Fawlty.Towers.S01E02.720p.BluRay.x264-SHORTBREHD.mkv',
              'size': 1168328357, 'crc': '156FDC0E'},
             'Fawlty.Towers.S01E03.720p.BluRay.x264-SHORTBREHD.mkv':
             {'release_name': 'Fawlty.Towers.S01E03.720p.BluRay.x264-SHORTBREHD',
              'file_name': 'Fawlty.Towers.S01E03.720p.BluRay.x264-SHORTBREHD.mkv',
              'size': 1559496197, 'crc': '5EAAAEC3'},
             'Fawlty.Towers.S01E04.720p.BluRay.x264-SHORTBREHD.mkv':
             {'release_name': 'Fawlty.Towers.S01E04.720p.BluRay.x264-SHORTBREHD',
              'file_name': 'Fawlty.Towers.S01E04.720p.BluRay.x264-SHORTBREHD.mkv',
              'size': 1560040644, 'crc': '4AAB7B8F'},
             'Fawlty.Towers.S01E05.720p.BluRay.x264-SHORTBREHD.mkv':
             {'release_name': 'Fawlty.Towers.S01E05.720p.BluRay.x264-SHORTBREHD',
              'file_name': 'Fawlty.Towers.S01E05.720p.BluRay.x264-SHORTBREHD.mkv',
              'size': 1559747905, 'crc': '40A2568C'},
             'Fawlty.Towers.S01E06.720p.BluRay.x264-SHORTBREHD.mkv':
             {'release_name': 'Fawlty.Towers.S01E06.720p.BluRay.x264-SHORTBREHD',
              'file_name': 'Fawlty.Towers.S01E06.720p.BluRay.x264-SHORTBREHD.mkv',
              'size': 1559556862, 'crc': '9FAB6E5F'}},
            id='Looking for multi-file release when original releases where single-file',
        ),

        pytest.param(
            'Drunk.History.S01.720p.HDTV.x264-2HD',
            {
                'drunk.history.s01e02.720p.hdtv.x264-2hd.mkv': {
                    'release_name': 'Drunk.History.S01E02.720p.HDTV.x264-2HD',
                    'file_name': 'drunk.history.s01e02.720p.hdtv.x264-2hd.mkv',
                    'size': 534364440, 'crc': 'F074DA2A',
                },
                'drunk.history.s01e03.720p.hdtv.x264-2hd.mkv': {
                    'release_name': 'Drunk.History.S01E03.720p.HDTV.x264-2HD',
                    'file_name': 'drunk.history.s01e03.720p.hdtv.x264-2hd.mkv',
                    'size': 525260014, 'crc': 'BBA99EED',
                },
            },
            id='Looking for incomplete season pack',
        ),

        pytest.param(
            'Bored.to.Death.S01E04.Deleted.Scene.720p.BluRay.x264-iNGOT.mkv',
            {'Bored.to.Death.S01E04.Deleted.Scene.720p.BluRay.x264-iNGOT.mkv':
             {'release_name': 'Bored.to.Death.S01.EXTRAS.720p.BluRay.x264-iNGOT',
              'file_name': 'Bored.to.Death.S01E04.Deleted.Scene.720p.BluRay.x264-iNGOT.mkv',
              'size': 138052914,
              'crc': 'E4E41C29'}},
            id='Looking for single episode from multi-file release',
        ),

        pytest.param(
            'Bored.to.Death.S01.Making.of.720p.BluRay.x264-iNGOT',
            {'Bored.to.Death.S01.Making.of.720p.BluRay.x264-iNGOT.mkv':
             {'release_name': 'Bored.to.Death.S01.EXTRAS.720p.BluRay.x264-iNGOT',
              'file_name': 'Bored.to.Death.S01.Making.of.720p.BluRay.x264-iNGOT.mkv',
              'size': 779447228,
              'crc': '8C8C0AE7'}},
            id='Looking for single non-episode file from multi-file release',
        ),

        pytest.param(
            'Foo.2015.1080p.Bluray.DD5.1.x264-NOTSCENEGROUP.mkv',
            {},
            id='Non-scene movie',
        ),

        pytest.param(
            'Foo.S01E03.1080p.Bluray.DD5.1.x264-NOTSCENEGROUP.mkv',
            {},
            id='Non-scene episode',
        ),

        pytest.param(
            'Foo.S03.1080p.Bluray.DD5.1.x264-NOTSCENEGROUP.mkv',
            {},
            id='Non-scene season pack',
        ),
    ),
)
@pytest.mark.asyncio
async def test_release_files(release_name, exp_return_value, multipredb, store_response):
    assert await multipredb.release_files(release_name) == exp_return_value


@pytest.mark.parametrize('parent_path', ('', 'path/to'))
@pytest.mark.parametrize(
    argnames='content_path, release_name, exp_exception',
    argvalues=(
        pytest.param(
            'Foo.2011.1080p.Bluray.DD5.1.x264-NOTASCENEGROUP.mkv',
            'Foo.2011.1080p.Bluray.DD5.1.x264-NOTASCENEGROUP',
            errors.SceneError('Not a scene release: Foo.2011.1080p.Bluray.DD5.1.x264-NOTASCENEGROUP'),
            id='[MOVIE] Non-scene release',
        ),

        pytest.param(
            'Dellamorte.Dellamore.1994.1080p.BluRay.x264-LiViDiTY',
            'Dellamorte.Dellamore.1994.1080p.BluRay.x264-LiViDiTY',
            None,
            id='[MOVIE] Properly named directory',
        ),

        pytest.param(
            f'Dellamorte.Dellamore.1994.1080p.BluRay.x264-LiViDiTY{os.sep}',
            'Dellamorte.Dellamore.1994.1080p.BluRay.x264-LiViDiTY',
            None,
            id='[MOVIE] Properly named directory with trailing os.sep',
        ),

        pytest.param(
            'Dellamorte.Dellamore.1994.1080p.BluRay.x264-LiViDiTY.mkv',
            'Dellamorte.Dellamore.1994.1080p.BluRay.x264-LiViDiTY',
            None,
            id='[MOVIE] Properly named file from abbreviated file name',
        ),

        pytest.param(
            'Dellamorte.Dellamore.1994.1080p.Blu-ray.x264-LiViDiTY',
            'Dellamorte.Dellamore.1994.1080p.BluRay.x264-LiViDiTY',
            errors.SceneRenamedError(original_name='Dellamorte.Dellamore.1994.1080p.BluRay.x264-LiViDiTY',
                                     existing_name='Dellamorte.Dellamore.1994.1080p.Blu-ray.x264-LiViDiTY'),
            id='[MOVIE] Renamed',
        ),

        pytest.param(
            'Dellamorte.Dellamore.1994.1080p.BluRay.x264-LiViDiTY/ly-dellmdm1080p.mkv',
            'Dellamorte.Dellamore.1994.1080p.BluRay.x264-LiViDiTY',
            None,
            id='[MOVIE] Abbreviated file in properly named directory',
        ),

        pytest.param(
            'Dellamorte.Dellamore.1994.1080p.Blu-ray.x264-LiViDiTY/ly-dellmdm1080p.mkv',
            'Dellamorte.Dellamore.1994.1080p.BluRay.x264-LiViDiTY',
            errors.SceneRenamedError(original_name='Dellamorte.Dellamore.1994.1080p.BluRay.x264-LiViDiTY/ly-dellmdm1080p.mkv',
                                     existing_name='Dellamorte.Dellamore.1994.1080p.Blu-ray.x264-LiViDiTY/ly-dellmdm1080p.mkv'),
            id='[MOVIE] Abbreviated file in renamed directory',
        ),

        pytest.param(
            'side.effects.2013.720p.bluray.x264-sparks.mkv',
            'Side.Effects.2013.720p.BluRay.x264-SPARKS',
            None,
            id='[MOVIE] File from release',
        ),

        pytest.param(
            'side.effects.2013.720p.bluray.x264-SPARKS.mkv',
            'Side.Effects.2013.720p.BluRay.x264-SPARKS',
            errors.SceneRenamedError(original_name='side.effects.2013.720p.bluray.x264-sparks.mkv',
                                     existing_name='side.effects.2013.720p.bluray.x264-SPARKS.mkv'),
            id='[MOVIE] Renamed file from release',
        ),

        pytest.param(
            'Friends.S10.1080p.BluRay.x264-TENEIGHTY/',
            'Friends.S10.1080p.BluRay.x264-TENEIGHTY',
            None,
            id='[SEASON] Abbreviated file names; properly named season directory',
        ),

        pytest.param(
            'Friends.S10.1080p.Blu-ray.x264-TENEIGHTY/',
            'Friends.S10.1080p.BluRay.x264-TENEIGHTY',
            errors.SceneRenamedError(original_name='Friends.S10.1080p.BluRay.x264-TENEIGHTY',
                                     existing_name='Friends.S10.1080p.Blu-ray.x264-TENEIGHTY'),
            id='[SEASON] Abbreviated file names; renamed season directory',
        ),

        pytest.param(
            'Fawlty.Towers.S01.720p.BluRay.x264-SHORTBREHD/',
            'Fawlty.Towers.S01.720p.BluRay.x264-SHORTBREHD',
            None,
            id='[SEASON] Normal file names; properly named season directory',
        ),

        pytest.param(
            'Fawlty.Towers.S01.720p.Blu-ray.x264-SHORTBREHD/',
            'Fawlty.Towers.S01.720p.BluRay.x264-SHORTBREHD',
            errors.SceneRenamedError(original_name='Fawlty.Towers.S01.720p.BluRay.x264-SHORTBREHD',
                                     existing_name='Fawlty.Towers.S01.720p.Blu-ray.x264-SHORTBREHD'),
            id='[SEASON] Normal file names; properly named season directory',
        ),

        pytest.param(
            'Friends.S10.1080p.BluRay.x264-TENEIGHTY/teneighty-friendss10e17e18.mkv',
            'Friends.S10E17E18.1080p.BluRay.x264-TENEIGHTY',
            None,
            id='[EPISODE] Abbreviated file name in properly named season directory',
        ),

        pytest.param(
            'Friends.S10.1080p.Blu-ray.x264-TENEIGHTY/teneighty-friendss10e17e18.mkv',
            'Friends.S10.1080p.BluRay.x264-TENEIGHTY',
            errors.SceneRenamedError(original_name='Friends.S10.1080p.BluRay.x264-TENEIGHTY/teneighty-friendss10e17e18.mkv',
                                     existing_name='Friends.S10.1080p.Blu-ray.x264-TENEIGHTY/teneighty-friendss10e17e18.mkv'),
            id='[EPISODE] Abbreviated file name in renamed season directory',
        ),

        pytest.param(
            'Friends.S10E17E18.1080p.BluRay.x264-TENEIGHTY/teneighty-friendss10e17e18.mkv',
            'Friends.S10E17E18.1080p.BluRay.x264-TENEIGHTY',
            None,
            id='[EPISODE] Abbreviated file name in properly named episode directory',
        ),

        pytest.param(
            'Friends.S10E17E18.1080p.Blu-ray.x264-TENEIGHTY/teneighty-friendss10e17e18.mkv',
            'Friends.S10E17E18.1080p.BluRay.x264-TENEIGHTY',
            errors.SceneRenamedError(original_name='Friends.S10E17E18.1080p.BluRay.x264-TENEIGHTY/teneighty-friendss10e17e18.mkv',
                                     existing_name='Friends.S10E17E18.1080p.Blu-ray.x264-TENEIGHTY/teneighty-friendss10e17e18.mkv'),
            id='[EPISODE] Abbreviated file name in renamed episode directory',
        ),

        pytest.param(
            'Friends.S10E17E18.1080p.BluRay.x264-TENEIGHTY/teneighty-friends.s10e17e18.mkv',
            'Friends.S10E17E18.1080p.BluRay.x264-TENEIGHTY',
            errors.SceneRenamedError(original_name='Friends.S10E17E18.1080p.BluRay.x264-TENEIGHTY/teneighty-friendss10e17e18.mkv',
                                     existing_name='Friends.S10E17E18.1080p.BluRay.x264-TENEIGHTY/teneighty-friends.s10e17e18.mkv'),
            id='[EPISODE] Renamed abbreviated file name in properly named episode directory',
        ),

        pytest.param(
            'Friends.S10E17E18.1080p.Blu-ray.x264-TENEIGHTY/teneighty-friends.s10e17e18.mkv',
            'Friends.S10E17E18.1080p.BluRay.x264-TENEIGHTY',
            errors.SceneRenamedError(original_name='Friends.S10E17E18.1080p.BluRay.x264-TENEIGHTY/teneighty-friendss10e17e18.mkv',
                                     existing_name='Friends.S10E17E18.1080p.Blu-ray.x264-TENEIGHTY/teneighty-friends.s10e17e18.mkv'),
            id='[EPISODE] Renamed abbreviated file name in renamed episode directory',
        ),

        pytest.param(
            'Fawlty.Towers.S01E06.720p.BluRay.x264-SHORTBREHD.mkv',
            'Fawlty.Towers.S01E06.720p.BluRay.x264-SHORTBREHD',
            None,
            id='[EPISODE] Normal file name',
        ),

        pytest.param(
            'Fawlty.Towers.S01E06.720p.Blu-ray.x264-SHORTBREHD.mkv',
            'Fawlty.Towers.S01E06.720p.BluRay.x264-SHORTBREHD',
            errors.SceneRenamedError(original_name='Fawlty.Towers.S01E06.720p.BluRay.x264-SHORTBREHD.mkv',
                                     existing_name='Fawlty.Towers.S01E06.720p.Blu-ray.x264-SHORTBREHD.mkv'),
            id='[EPISODE] Normal renamed file name',
        ),

        pytest.param(
            'Bored.to.Death.S01.EXTRAS.720p.BluRay.x264-iNGOT',
            'Bored.to.Death.S01.EXTRAS.720p.BluRay.x264-iNGOT',
            None,
            id='[EXTRAS] Multi-file release',
        ),

        pytest.param(
            'Bored.to.Death.S01.Extras.720p.BluRay.x264-iNGOT',
            'Bored.to.Death.S01.EXTRAS.720p.BluRay.x264-iNGOT',
            errors.SceneRenamedError(original_name='Bored.to.Death.S01.EXTRAS.720p.BluRay.x264-iNGOT',
                                     existing_name='Bored.to.Death.S01.Extras.720p.BluRay.x264-iNGOT'),
            id='[EXTRAS] Renamed multi-file release',
        ),

        pytest.param(
            'Bored.to.Death.S01.Making.of.720p.BluRay.x264-iNGOT.mkv',
            'Bored.to.Death.S01.EXTRAS.720p.BluRay.x264-iNGOT',
            None,
            id='[EXTRAS] Single file from multi-file release',
        ),

        pytest.param(
            'foo',
            'Wrecked.2011.DiRFiX.LIMITED.FRENCH.720p.BluRay.X264-LOST',
            errors.SceneRenamedError(original_name='Wrecked.2011.DiRFiX.LIMITED.FRENCH.720p.BluRay.X264-LOST',
                                     existing_name='foo'),
            id='Empty release',
        ),
    ),
)
@pytest.mark.asyncio
async def test_verify_release_name(content_path, release_name, exp_exception, parent_path, multipredb, store_response):
    full_content_path = os.path.join(parent_path, content_path)
    if exp_exception:
        with pytest.raises(type(exp_exception), match=rf'^{re.escape(str(exp_exception))}$'):
            await multipredb.verify_release_name(full_content_path, release_name)
    else:
        await multipredb.verify_release_name(full_content_path, release_name)


@pytest.mark.parametrize(
    argnames='content_path, release_name, file_sizes, exp_exceptions',
    argvalues=(
        pytest.param(
            'path/to/Foo.2011.1080p.Bluray.DD5.1.x264-NOTASCENEGROUP.mkv',
            'Foo.2011.1080p.Bluray.DD5.1.x264-NOTASCENEGROUP',
            (('path/to/Foo.2011.1080p.Bluray.DD5.1.x264-NOTASCENEGROUP.mkv', 7793811601),),
            (errors.SceneError('Not a scene release: Foo.2011.1080p.Bluray.DD5.1.x264-NOTASCENEGROUP'),),
            id='Non-scene release',
        ),

        pytest.param(
            'path/to/Dellamorte.Dellamore.1994.1080p.BluRay.x264-LiViDiTY/ly-dellmdm1080p.mkv',
            'Dellamorte.Dellamore.1994.1080p.BluRay.x264-LiViDiTY',
            (('path/to/Dellamorte.Dellamore.1994.1080p.BluRay.x264-LiViDiTY/ly-dellmdm1080p.mkv', 7044061767),),
            (),
            id='Abbreviated file in properly named directory with correct size',
        ),

        pytest.param(
            'path/to/Dellamorte.Dellamore.1994.1080p.BluRay.x264-LiViDiTY/ly-dellmdm1080p.mkv',
            'Dellamorte.Dellamore.1994.1080p.BluRay.x264-LiViDiTY',
            (('path/to/Dellamorte.Dellamore.1994.1080p.BluRay.x264-LiViDiTY/ly-dellmdm1080p.mkv', 123),),
            (errors.SceneFileSizeError(filename='ly-dellmdm1080p.mkv', original_size=7044061767, existing_size=123),),
            id='Abbreviated file in properly named directory with wrong size',
        ),

        pytest.param(
            'path/to/ly-dellmdm1080p.mkv',
            'Dellamorte.Dellamore.1994.1080p.BluRay.x264-LiViDiTY',
            (('path/to/ly-dellmdm1080p.mkv', 7044061767),),
            (),
            id='Abbreviated file in non-release directory with correct size',
        ),

        pytest.param(
            'path/to/ly-dellmdm1080p.mkv',
            'Dellamorte.Dellamore.1994.1080p.BluRay.x264-LiViDiTY',
            (('path/to/ly-dellmdm1080p.mkv', 123),),
            (errors.SceneFileSizeError(filename='ly-dellmdm1080p.mkv', original_size=7044061767, existing_size=123),),
            id='Abbreviated file in non-release directory with wrong size',
        ),

        pytest.param(
            'path/to/Dellamorte.Dellamore.1994.1080p.BluRay.x264-LiViDiTY.mkv',
            'Dellamorte.Dellamore.1994.1080p.BluRay.x264-LiViDiTY',
            (('path/to/Dellamorte.Dellamore.1994.1080p.BluRay.x264-LiViDiTY.mkv', 7044061767),),
            (),
            id='Properly named file with correct size',
        ),

        pytest.param(
            'path/to/Dellamorte.Dellamore.1994.1080p.BluRay.x264-LiViDiTY.mkv',
            'Dellamorte.Dellamore.1994.1080p.BluRay.x264-LiViDiTY',
            (('path/to/Dellamorte.Dellamore.1994.1080p.BluRay.x264-LiViDiTY.mkv', 123),),
            (errors.SceneFileSizeError(filename='Dellamorte.Dellamore.1994.1080p.BluRay.x264-LiViDiTY.mkv',
                                       original_size=7044061767, existing_size=123),),
            id='Properly named file with wrong size',
        ),

        pytest.param(
            'path/to/dellamorte.dellamore.1994.1080p.blu-ray.x264-lividity.mkv',
            'Dellamorte.Dellamore.1994.1080p.BluRay.x264-LiViDiTY',
            (('path/to/dellamorte.dellamore.1994.1080p.blu-ray.x264-lividity.mkv', 7044061767),),
            (),
            id='Renamed file with correct size',
        ),

        pytest.param(
            'path/to/dellamorte.dellamore.1994.1080p.bluray.x264-lividity.mkv',
            'Dellamorte.Dellamore.1994.1080p.BluRay.x264-LiViDiTY',
            (('path/to/dellamorte.dellamore.1994.1080p.bluray.x264-lividity.mkv', 123),),
            (errors.SceneFileSizeError(filename='dellamorte.dellamore.1994.1080p.bluray.x264-lividity.mkv',
                                       original_size=7044061767, existing_size=123),),
            id='Renamed file with wrong size',
        ),

        pytest.param(
            'path/to/The.Fall.S02.720p.BluRay.x264-7SinS',
            'The.Fall.S02.720p.BluRay.x264-7SinS',
            (('path/to/The.Fall.S02.720p.BluRay.x264-7SinS/7s-tf-s02e01-720p.mkv', 2340971701),
             ('path/to/The.Fall.S02.720p.BluRay.x264-7SinS/7s-tf-s02e02-720p.mkv', 2345872861),
             ('path/to/The.Fall.S02.720p.BluRay.x264-7SinS/7s-tf-s02e03-720p.mkv', 2345958712),
             ('path/to/The.Fall.S02.720p.BluRay.x264-7SinS/7s-tf-s02e04-720p.mkv', 2346136836),
             ('path/to/The.Fall.S02.720p.BluRay.x264-7SinS/7s-tf-s02e05-720p.mkv', 123),  # No info about this one
             ('path/to/The.Fall.S02.720p.BluRay.x264-7SinS/7s-tf-s02e06-720p.mkv', 3517999979)),
            (errors.SceneFileSizeError(filename='7s-tf-s02e02-720p.mkv', original_size=2345872860, existing_size=2345872861),
             errors.SceneMissingInfoError('7s-tf-s02e05-720p.mkv'),
             errors.SceneFileSizeError(filename='7s-tf-s02e06-720p.mkv', original_size=3517999978, existing_size=3517999979)),
            id='Missing scene release info / wrong episode order in release',
        ),

        pytest.param(
            'path/to/Bored.to.Death.S01.EXTRAS.720p.BluRay.x264-iNGOT',
            'Bored.to.Death.S01.EXTRAS.720p.BluRay.x264-iNGOT',
            (('path/to/Bored.to.Death.S01.EXTRAS.720p.BluRay.x264-iNGOT/Bored.to.Death.S01.Jonathan.Amess.Brooklyn.720p.BluRay.x264-iNGOT.mkv', 518652629),
             ('path/to/Bored.to.Death.S01.EXTRAS.720p.BluRay.x264-iNGOT/Bored.to.Death.S01.Making.of.720p.BluRay.x264-iNGOT.mkv', 779447228),
             ('path/to/Bored.to.Death.S01.EXTRAS.720p.BluRay.x264-iNGOT/Bored.to.Death.S01E03.Deleted.Scene.720p.BluRay.x264-iNGOT.mkv', 30779540),
             ('path/to/Bored.to.Death.S01.EXTRAS.720p.BluRay.x264-iNGOT/Bored.to.Death.S01E04.Deleted.Scene.720p.BluRay.x264-iNGOT.mkv', 138052913),
             ('path/to/Bored.to.Death.S01.EXTRAS.720p.BluRay.x264-iNGOT/Bored.to.Death.S01E08.Deleted.Scene.1.720p.BluRay.x264-iNGOT.mkv', 68498554),
             ('path/to/Bored.to.Death.S01.EXTRAS.720p.BluRay.x264-iNGOT/Bored.to.Death.S01E08.Deleted.Scene.2.720p.BluRay.x264-iNGOT.mkv', 45583011),),
            (errors.SceneFileSizeError(filename='Bored.to.Death.S01E04.Deleted.Scene.720p.BluRay.x264-iNGOT.mkv',
                                       original_size=138052914, existing_size=138052913),),
            id='Multi-file release with one incorrectly sized file',
        ),

        pytest.param(
            'path/to/Bored.to.Death.S01.Making.of.720p.BluRay.x264-iNGOT.mkv',
            'Bored.to.Death.S01.EXTRAS.720p.BluRay.x264-iNGOT',
            (('path/to/Bored.to.Death.S01.Making.of.720p.BluRay.x264-iNGOT.mkv', 779447228),),
            (),
            id='Single file with correct size from multi-file release',
        ),

        pytest.param(
            'path/to/Bored.to.Death.S01.Making.of.720p.BluRay.x264-iNGOT.mkv',
            'Bored.to.Death.S01.EXTRAS.720p.BluRay.x264-iNGOT',
            (('path/to/Bored.to.Death.S01.Making.of.720p.BluRay.x264-iNGOT.mkv', 779447229),),
            (errors.SceneFileSizeError(filename='Bored.to.Death.S01.Making.of.720p.BluRay.x264-iNGOT.mkv',
                                       original_size=779447228, existing_size=779447229),),
            id='Single file with wrong size from multi-file release',
        ),

        pytest.param(
            'path/to/Bored.to.Death.S01.MAKING.OF.720p.BluRay.x264-iNGOT.mkv',
            'Bored.to.Death.S01.EXTRAS.720p.BluRay.x264-iNGOT',
            (('path/to/Bored.to.Death.S01.MAKING.OF.720p.BluRay.x264-iNGOT.mkv', 779447228),),
            (errors.SceneMissingInfoError('Bored.to.Death.S01.MAKING.OF.720p.BluRay.x264-iNGOT.mkv'),),
            id='Renamed single file from multi-file release',
        ),
    ),
)
@pytest.mark.asyncio
async def test_verify_release_files(content_path, release_name, file_sizes, exp_exceptions, multipredb, store_response, tmp_path):
    for filepath, filesize in file_sizes:
        parent = tmp_path / os.path.dirname(filepath)
        if not parent.exists():
            parent.mkdir(parents=True)
        with (tmp_path / filepath).open('wb') as f:
            f.truncate(filesize)  # Sparse file
        assert os.path.getsize(tmp_path / filepath) == filesize

    exceptions = await multipredb.verify_release_files(str(tmp_path / content_path), release_name)
    assert exceptions == exp_exceptions


@pytest.mark.parametrize(
    argnames='content_path, release_name, file_sizes, exp_scene_check_result, exp_exceptions',
    argvalues=(
        pytest.param(
            'path/to/Foo.2011.1080p.Bluray.DD5.1.x264-NOTASCENEGROUP.mkv',
            'Foo.2011.1080p.Bluray.DD5.1.x264-NOTASCENEGROUP',
            (('path/to/Foo.2011.1080p.Bluray.DD5.1.x264-NOTASCENEGROUP.mkv', 7793811601),),
            SceneCheckResult.false,
            (),
            id='Non-scene release',
        ),

        pytest.param(
            'path/to/Dellamorte.Dellamore.1994.1080p.BluRay.x264-LiViDiTY/ly-dellmdm1080p.mkv',
            'Dellamorte.Dellamore.1994.1080p.BluRay.x264-LiViDiTY',
            (('path/to/Dellamorte.Dellamore.1994.1080p.BluRay.x264-LiViDiTY/ly-dellmdm1080p.mkv', 7044061767),),
            SceneCheckResult.true,
            (),
            id='Abbreviated file in properly named directory with correct size',
        ),

        pytest.param(
            'path/to/Dellamorte.Dellamore.1994.1080p.BluRay.x264-LiViDiTY/ly-dellmdm1080p.mkv',
            'Dellamorte.Dellamore.1994.1080p.BluRay.x264-LiViDiTY',
            (('path/to/Dellamorte.Dellamore.1994.1080p.BluRay.x264-LiViDiTY/ly-dellmdm1080p.mkv', 123),),
            SceneCheckResult.true,
            (errors.SceneFileSizeError(filename='ly-dellmdm1080p.mkv', original_size=7044061767, existing_size=123),),
            id='Abbreviated file in properly named directory with wrong size',
        ),

        pytest.param(
            'path/to/ly-dellmdm1080p.mkv',
            'Dellamorte.Dellamore.1994.1080p.BluRay.x264-LiViDiTY',
            (('path/to/ly-dellmdm1080p.mkv', 7044061767),),
            SceneCheckResult.true,
            (errors.SceneRenamedError(original_name='Dellamorte.Dellamore.1994.1080p.BluRay.x264-LiViDiTY/ly-dellmdm1080p.mkv',
                                      existing_name='to/ly-dellmdm1080p.mkv'),),
            id='Abbreviated file in non-release directory with correct size',
        ),

        pytest.param(
            'path/to/ly-dellmdm1080p.mkv',
            'Dellamorte.Dellamore.1994.1080p.BluRay.x264-LiViDiTY',
            (('path/to/ly-dellmdm1080p.mkv', 123),),
            SceneCheckResult.true,
            (errors.SceneRenamedError(original_name='Dellamorte.Dellamore.1994.1080p.BluRay.x264-LiViDiTY/ly-dellmdm1080p.mkv',
                                      existing_name='to/ly-dellmdm1080p.mkv'),
             errors.SceneFileSizeError(filename='ly-dellmdm1080p.mkv', original_size=7044061767, existing_size=123)),
            id='Abbreviated file in non-release directory with wrong size',
        ),

        pytest.param(
            'path/to/Dellamorte.Dellamore.1994.1080p.BluRay.x264-LiViDiTY.mkv',
            'Dellamorte.Dellamore.1994.1080p.BluRay.x264-LiViDiTY',
            (('path/to/Dellamorte.Dellamore.1994.1080p.BluRay.x264-LiViDiTY.mkv', 7044061767),),
            SceneCheckResult.true,
            (),
            id='Properly named file with correct size',
        ),

        pytest.param(
            'path/to/Dellamorte.Dellamore.1994.1080p.BluRay.x264-LiViDiTY.mkv',
            'Dellamorte.Dellamore.1994.1080p.BluRay.x264-LiViDiTY',
            (('path/to/Dellamorte.Dellamore.1994.1080p.BluRay.x264-LiViDiTY.mkv', 123),),
            SceneCheckResult.true,
            (errors.SceneFileSizeError(filename='Dellamorte.Dellamore.1994.1080p.BluRay.x264-LiViDiTY.mkv',
                                       original_size=7044061767, existing_size=123),),
            id='Properly named file with wrong size',
        ),

        pytest.param(
            'path/to/dellamorte.dellamore.1994.1080p.blu-ray.x264-lividity.mkv',
            'Dellamorte.Dellamore.1994.1080p.BluRay.x264-LiViDiTY',
            (('path/to/dellamorte.dellamore.1994.1080p.blu-ray.x264-lividity.mkv', 7044061767),),
            SceneCheckResult.true,
            (errors.SceneRenamedError(original_name='ly-dellmdm1080p.mkv',
                                      existing_name='dellamorte.dellamore.1994.1080p.blu-ray.x264-lividity.mkv'),),
            id='Renamed file with correct size',
        ),

        pytest.param(
            'path/to/dellamorte.dellamore.1994.1080p.bluray.x264-lividity.mkv',
            'Dellamorte.Dellamore.1994.1080p.BluRay.x264-LiViDiTY',
            (('path/to/dellamorte.dellamore.1994.1080p.bluray.x264-lividity.mkv', 123),),
            SceneCheckResult.true,
            (errors.SceneRenamedError(original_name='ly-dellmdm1080p.mkv',
                                      existing_name='dellamorte.dellamore.1994.1080p.bluray.x264-lividity.mkv'),
             errors.SceneFileSizeError(filename='dellamorte.dellamore.1994.1080p.bluray.x264-lividity.mkv',
                                       original_size=7044061767, existing_size=123),),
            id='Renamed file with wrong size',
        ),

        pytest.param(
            'path/to/The.Fall.S02.720p.BluRay.x264-7SinS',
            'The.Fall.S02.720p.BluRay.x264-7SinS',
            (('path/to/The.Fall.S02.720p.BluRay.x264-7SinS/7s-tf-s02e01-720p.mkv', 2340971701),
             ('path/to/The.Fall.S02.720p.BluRay.x264-7SinS/7s-tf-s02e02-720p.mkv', 2345872861),
             ('path/to/The.Fall.S02.720p.BluRay.x264-7SinS/7s-tf-s02e03-720p.mkv', 2345958712),
             ('path/to/The.Fall.S02.720p.BluRay.x264-7SinS/7s-tf-s02e04-720p.mkv', 2346136836),
             ('path/to/The.Fall.S02.720p.BluRay.x264-7SinS/7s-tf-s02e05-720p.mkv', 123),  # No info about this one
             ('path/to/The.Fall.S02.720p.BluRay.x264-7SinS/7s-tf-s02e06-720p.mkv', 3517999979)),
            SceneCheckResult.true,
            (errors.SceneFileSizeError(filename='7s-tf-s02e02-720p.mkv', original_size=2345872860, existing_size=2345872861),
             errors.SceneMissingInfoError('7s-tf-s02e05-720p.mkv'),
             errors.SceneFileSizeError(filename='7s-tf-s02e06-720p.mkv', original_size=3517999978, existing_size=3517999979)),
            id='Missing scene release info / wrong episode order in release',
        ),

        pytest.param(
            'path/to/Bored.to.Death.S01.EXTRAS.720p.BluRay.x264-iNGOT',
            'Bored.to.Death.S01.EXTRAS.720p.BluRay.x264-iNGOT',
            (('path/to/Bored.to.Death.S01.EXTRAS.720p.BluRay.x264-iNGOT/Bored.to.Death.S01.Jonathan.Amess.Brooklyn.720p.BluRay.x264-iNGOT.mkv', 518652629),
             ('path/to/Bored.to.Death.S01.EXTRAS.720p.BluRay.x264-iNGOT/Bored.to.Death.S01.Making.of.720p.BluRay.x264-iNGOT.mkv', 779447228),
             ('path/to/Bored.to.Death.S01.EXTRAS.720p.BluRay.x264-iNGOT/Bored.to.Death.S01E03.Deleted.Scene.720p.BluRay.x264-iNGOT.mkv', 30779540),
             ('path/to/Bored.to.Death.S01.EXTRAS.720p.BluRay.x264-iNGOT/Bored.to.Death.S01E04.Deleted.Scene.720p.BluRay.x264-iNGOT.mkv', 138052913),
             ('path/to/Bored.to.Death.S01.EXTRAS.720p.BluRay.x264-iNGOT/Bored.to.Death.S01E08.Deleted.Scene.1.720p.BluRay.x264-iNGOT.mkv', 68498554),
             ('path/to/Bored.to.Death.S01.EXTRAS.720p.BluRay.x264-iNGOT/Bored.to.Death.S01E08.Deleted.Scene.2.720p.BluRay.x264-iNGOT.mkv', 45583011),),
            SceneCheckResult.true,
            (errors.SceneFileSizeError(filename='Bored.to.Death.S01E04.Deleted.Scene.720p.BluRay.x264-iNGOT.mkv',
                                       original_size=138052914, existing_size=138052913),),
            id='Multi-file release with one incorrectly sized file',
        ),

        pytest.param(
            'path/to/Bored.to.Death.S01.Making.of.720p.BluRay.x264-iNGOT.mkv',
            'Bored.to.Death.S01.EXTRAS.720p.BluRay.x264-iNGOT',
            (('path/to/Bored.to.Death.S01.Making.of.720p.BluRay.x264-iNGOT.mkv', 779447228),),
            SceneCheckResult.true,
            (),
            id='Single file with correct size from multi-file release',
        ),

        pytest.param(
            'path/to/Bored.to.Death.S01.Making.of.720p.BluRay.x264-iNGOT.mkv',
            'Bored.to.Death.S01.EXTRAS.720p.BluRay.x264-iNGOT',
            (('path/to/Bored.to.Death.S01.Making.of.720p.BluRay.x264-iNGOT.mkv', 779447229),),
            SceneCheckResult.true,
            (errors.SceneFileSizeError(filename='Bored.to.Death.S01.Making.of.720p.BluRay.x264-iNGOT.mkv',
                                       original_size=779447228, existing_size=779447229),),
            id='Single file with wrong size from multi-file release',
        ),

        pytest.param(
            'path/to/Bored.to.Death.S01.MAKING.OF.720p.BluRay.x264-iNGOT.mkv',
            'Bored.to.Death.S01.EXTRAS.720p.BluRay.x264-iNGOT',
            (('path/to/Bored.to.Death.S01.MAKING.OF.720p.BluRay.x264-iNGOT.mkv', 779447228),),
            SceneCheckResult.true,
            (errors.SceneRenamedError(original_name='Bored.to.Death.S01.Making.of.720p.BluRay.x264-iNGOT.mkv',
                                      existing_name='Bored.to.Death.S01.MAKING.OF.720p.BluRay.x264-iNGOT.mkv'),
             errors.SceneMissingInfoError('Bored.to.Death.S01.MAKING.OF.720p.BluRay.x264-iNGOT.mkv'),),
            id='Renamed single file from multi-file release',
        ),
    ),
)
@pytest.mark.asyncio
async def test_verify_release_with_release_name(content_path, release_name, file_sizes, exp_scene_check_result, exp_exceptions,
                                                multipredb, store_response, tmp_path):
    for filepath, filesize in file_sizes:
        parent = tmp_path / os.path.dirname(filepath)
        if not parent.exists():
            parent.mkdir(parents=True)
        with (tmp_path / filepath).open('wb') as f:
            f.truncate(filesize)  # Sparse file
        assert os.path.getsize(tmp_path / filepath) == filesize

    result, exceptions = await multipredb.verify_release(str(tmp_path / content_path), release_name)
    assert result is exp_scene_check_result
    assert exceptions == exp_exceptions


@pytest.mark.parametrize(
    argnames='content_path, file_sizes, exp_scene_check_result, exp_exceptions',
    argvalues=(
        pytest.param(
            'path/to/Foo.2011.1080p.Bluray.DD5.1.x264-NOTASCENEGROUP.mkv',
            (('path/to/Foo.2011.1080p.Bluray.DD5.1.x264-NOTASCENEGROUP.mkv', 7793811601),),
            SceneCheckResult.false,
            (),
            id='Non-scene release',
        ),

        pytest.param(
            'path/to/Dellamorte.Dellamore.1994.1080p.BluRay.x264-LiViDiTY/ly-dellmdm1080p.mkv',
            (('path/to/Dellamorte.Dellamore.1994.1080p.BluRay.x264-LiViDiTY/ly-dellmdm1080p.mkv', 7044061767),),
            SceneCheckResult.true,
            (),
            id='Abbreviated file in properly named directory with correct size',
        ),

        pytest.param(
            'path/to/Dellamorte.Dellamore.1994.1080p.BluRay.x264-LiViDiTY/ly-dellmdm1080p.mkv',
            (('path/to/Dellamorte.Dellamore.1994.1080p.BluRay.x264-LiViDiTY/ly-dellmdm1080p.mkv', 123),),
            SceneCheckResult.true,
            (errors.SceneFileSizeError(filename='ly-dellmdm1080p.mkv', original_size=7044061767, existing_size=123),),
            id='Abbreviated file in properly named directory with wrong size',
        ),

        pytest.param(
            'path/to/Foo.2011.1080p.BluRay.x264-NOTASCENEGROUP/ly-dellmdm1080p.mkv',
            (('path/to/Foo.2011.1080p.BluRay.x264-NOTASCENEGROUP/ly-dellmdm1080p.mkv', 7044061767),),
            SceneCheckResult.false,
            (),
            id='Abbreviated file in non-release directory with correct size',
        ),

        pytest.param(
            'path/to/Foo.2011.1080p.BluRay.x264-NOTASCENEGROUP/ly-dellmdm1080p.mkv',
            (('path/to/Foo.2011.1080p.BluRay.x264-NOTASCENEGROUP/ly-dellmdm1080p.mkv', 123),),
            SceneCheckResult.false,
            (),
            id='Abbreviated file in non-release directory with wrong size',
        ),

        pytest.param(
            'path/to/Dellamorte.Dellamore.1994.1080p.BluRay.x264-LiViDiTY.mkv',
            (('path/to/Dellamorte.Dellamore.1994.1080p.BluRay.x264-LiViDiTY.mkv', 7044061767),),
            SceneCheckResult.true,
            (),
            id='Properly named file with correct size',
        ),

        pytest.param(
            'path/to/Dellamorte.Dellamore.1994.1080p.BluRay.x264-LiViDiTY.mkv',
            (('path/to/Dellamorte.Dellamore.1994.1080p.BluRay.x264-LiViDiTY.mkv', 123),),
            SceneCheckResult.true,
            (errors.SceneFileSizeError(filename='Dellamorte.Dellamore.1994.1080p.BluRay.x264-LiViDiTY.mkv',
                                       original_size=7044061767, existing_size=123),),
            id='Properly named file with wrong size',
        ),

        pytest.param(
            'path/to/dellamorte.dellamore.1994.1080p.blu-ray.x264-lividity.mkv',
            (('path/to/dellamorte.dellamore.1994.1080p.blu-ray.x264-lividity.mkv', 7044061767),),
            SceneCheckResult.true,
            (errors.SceneRenamedError(original_name='ly-dellmdm1080p.mkv',
                                      existing_name='dellamorte.dellamore.1994.1080p.blu-ray.x264-lividity.mkv'),),
            id='Renamed file with correct size',
        ),

        pytest.param(
            'path/to/dellamorte.dellamore.1994.1080p.bluray.x264-lividity.mkv',
            (('path/to/dellamorte.dellamore.1994.1080p.bluray.x264-lividity.mkv', 123),),
            SceneCheckResult.true,
            (errors.SceneRenamedError(original_name='ly-dellmdm1080p.mkv',
                                      existing_name='dellamorte.dellamore.1994.1080p.bluray.x264-lividity.mkv'),
             errors.SceneFileSizeError(filename='dellamorte.dellamore.1994.1080p.bluray.x264-lividity.mkv',
                                       original_size=7044061767, existing_size=123),),
            id='Renamed file with wrong size',
        ),

        pytest.param(
            'path/to/The.Fall.S02.720p.BluRay.x264-7SinS',
            (('path/to/The.Fall.S02.720p.BluRay.x264-7SinS/7s-tf-s02e01-720p.mkv', 2340971701),
             ('path/to/The.Fall.S02.720p.BluRay.x264-7SinS/7s-tf-s02e02-720p.mkv', 2345872861),
             ('path/to/The.Fall.S02.720p.BluRay.x264-7SinS/7s-tf-s02e03-720p.mkv', 2345958712),
             ('path/to/The.Fall.S02.720p.BluRay.x264-7SinS/7s-tf-s02e04-720p.mkv', 2346136836),
             ('path/to/The.Fall.S02.720p.BluRay.x264-7SinS/7s-tf-s02e05-720p.mkv', 123),  # No info about this one
             ('path/to/The.Fall.S02.720p.BluRay.x264-7SinS/7s-tf-s02e06-720p.mkv', 3517999979)),
            SceneCheckResult.unknown,
            (errors.SceneFileSizeError(filename='7s-tf-s02e02-720p.mkv', original_size=2345872860, existing_size=2345872861),
             errors.SceneFileSizeError(filename='7s-tf-s02e06-720p.mkv', original_size=3517999978, existing_size=3517999979)),
            id='Missing scene release info / wrong episode order in release',
        ),

        pytest.param(
            'path/to/Bored.to.Death.S01.EXTRAS.720p.BluRay.x264-iNGOT',
            (('path/to/Bored.to.Death.S01.EXTRAS.720p.BluRay.x264-iNGOT/Bored.to.Death.S01.Jonathan.Amess.Brooklyn.720p.BluRay.x264-iNGOT.mkv', 518652629),
             ('path/to/Bored.to.Death.S01.EXTRAS.720p.BluRay.x264-iNGOT/Bored.to.Death.S01.Making.of.720p.BluRay.x264-iNGOT.mkv', 779447228),
             ('path/to/Bored.to.Death.S01.EXTRAS.720p.BluRay.x264-iNGOT/Bored.to.Death.S01E03.Deleted.Scene.720p.BluRay.x264-iNGOT.mkv', 30779540),
             ('path/to/Bored.to.Death.S01.EXTRAS.720p.BluRay.x264-iNGOT/Bored.to.Death.S01E04.Deleted.Scene.720p.BluRay.x264-iNGOT.mkv', 138052913),
             ('path/to/Bored.to.Death.S01.EXTRAS.720p.BluRay.x264-iNGOT/Bored.to.Death.S01E08.Deleted.Scene.1.720p.BluRay.x264-iNGOT.mkv', 68498554),
             ('path/to/Bored.to.Death.S01.EXTRAS.720p.BluRay.x264-iNGOT/Bored.to.Death.S01E08.Deleted.Scene.2.720p.BluRay.x264-iNGOT.mkv', 45583011),),
            SceneCheckResult.unknown,
            (),
            id='Multi-file release with one incorrectly sized file',
        ),

        pytest.param(
            'path/to/Bored.to.Death.S01.Making.of.720p.BluRay.x264-iNGOT.mkv',
            (('path/to/Bored.to.Death.S01.Making.of.720p.BluRay.x264-iNGOT.mkv', 779447228),),
            SceneCheckResult.true,
            (),
            id='Single file with correct size from multi-file release',
        ),

        pytest.param(
            'path/to/Bored.to.Death.S01.Making.of.720p.BluRay.x264-iNGOT.mkv',
            (('path/to/Bored.to.Death.S01.Making.of.720p.BluRay.x264-iNGOT.mkv', 779447229),),
            SceneCheckResult.true,
            (errors.SceneFileSizeError(filename='Bored.to.Death.S01.Making.of.720p.BluRay.x264-iNGOT.mkv',
                                       original_size=779447228, existing_size=779447229),),
            id='Single file with wrong size from multi-file release',
        ),

        pytest.param(
            'path/to/Bored.to.Death.S01.MAKING.OF.720p.BluRay.x264-iNGOT.mkv',
            (('path/to/Bored.to.Death.S01.MAKING.OF.720p.BluRay.x264-iNGOT.mkv', 779447228),),
            SceneCheckResult.true,
            (errors.SceneRenamedError(original_name='Bored.to.Death.S01.Making.of.720p.BluRay.x264-iNGOT.mkv',
                                      existing_name='Bored.to.Death.S01.MAKING.OF.720p.BluRay.x264-iNGOT.mkv'),
             errors.SceneMissingInfoError('Bored.to.Death.S01.MAKING.OF.720p.BluRay.x264-iNGOT.mkv'),),
            id='Renamed single file from multi-file release',
        ),
    ),
)
@pytest.mark.asyncio
async def test_verify_release_without_release_name(content_path, file_sizes, exp_scene_check_result, exp_exceptions,
                                                   multipredb, store_response, tmp_path):
    for filepath, filesize in file_sizes:
        parent = tmp_path / os.path.dirname(filepath)
        if not parent.exists():
            parent.mkdir(parents=True)
        with (tmp_path / filepath).open('wb') as f:
            f.truncate(filesize)  # Sparse file
        assert os.path.getsize(tmp_path / filepath) == filesize

    result, exceptions = await multipredb.verify_release(str(tmp_path / content_path))
    assert result is exp_scene_check_result
    assert exceptions == exp_exceptions



"""
Wrapper for ``mediainfo`` command
"""

import asyncio
import os

from .. import errors, utils
from . import JobBase

import logging  # isort:skip
_log = logging.getLogger(__name__)


class MediainfoJob(JobBase):
    """
    Get output from ``mediainfo`` command

    See :func:`.utils.video.mediainfo` for more information.
    """

    name = 'mediainfo'
    label = 'Mediainfo'

    # Don't show mediainfo output in TUI. It is printed out to stdout if this is
    # the only/final job.
    hidden = True

    @property
    def cache_id(self):
        """Final segment of `content_path` and `from_all_videos` argument"""
        return (
            utils.fs.basename(self._content_path),
            'from_all_videos' if self._from_all_videos else 'from_first_video',
            [
                # `re.Pattern` objects or glob strings
                (exclude.pattern if hasattr(exclude, 'pattern') else str(exclude))
                for exclude in self._exclude_files
            ],
        )

    def initialize(self, *, content_path, from_all_videos=False, exclude_files=()):
        """
        Set internal state

        :param content_path: Path to video file or directory that contains a
            video file
        :param bool from_all_videos: Whether to get ``mediainfo`` output from
            each video file or only from the first video
        :param exclude_files: Sequence of glob patterns (:class:`str`) and
            :class:`re.Pattern` objects (return value from :func:`re.compile`)
            that are matched against the relative path beneath `content_path`

            Glob patterns are matched case-insensitively.

            .. note:: Non-video files and stuff like `Sample.mkv` are always
                      excluded by :func:`.video.filter_main_videos`.
        """
        self._content_path = content_path
        self._from_all_videos = from_all_videos
        self._exclude_files = exclude_files
        self._mediainfos_by_file = {}
        self.signal.add('mediainfo', record=True)
        self.signal.register('mediainfo', self._store_mediainfos_by_file)

    def execute(self):
        """Call ``mediainfo`` in an asynchronous thread"""
        self.attach_task(
            coro=self._generate_mediainfos(),
            finish_when_done=True,
        )

    async def _generate_mediainfos(self):
        # Exclude files specified by the user
        video_filepaths = utils.torrent.filter_files(
            content_path=self._content_path,
            exclude=self._exclude_files,
        )
        # Exclude irrelevant files (non-videos, sample.mkv, etc)
        video_filepaths = utils.video.filter_main_videos(video_filepaths)
        for video_filepath in video_filepaths:
            # DVD
            if utils.fs.dirname(video_filepath).endswith(f'{os.sep}VIDEO_TS'):
                ifo_filepath = utils.video.vob2ifo(video_filepath)
                vob_filepath = video_filepath
                mediainfos_or_exceptions = await asyncio.gather(
                    utils.run_async(utils.video.mediainfo, ifo_filepath),
                    utils.run_async(utils.video.mediainfo, vob_filepath),
                    return_exceptions=True,
                )
                assert len(mediainfos_or_exceptions) == len((ifo_filepath, vob_filepath))
                for filepath, mi_or_exc in zip((ifo_filepath, vob_filepath), mediainfos_or_exceptions):
                    if isinstance(mi_or_exc, errors.ContentError):
                        self.error(mi_or_exc)
                    elif isinstance(mi_or_exc, Exception):
                        raise mi_or_exc
                    else:
                        self.send(mi_or_exc)
                        self.signal.emit('mediainfo', filepath, mi_or_exc)

            # Normal video file(s)
            else:
                try:
                    mi = await utils.run_async(utils.video.mediainfo, video_filepath)
                except errors.ContentError as e:
                    self.error(e)
                else:
                    self.send(mi)
                    self.signal.emit('mediainfo', video_filepath, mi)

            if not self._from_all_videos:
                break

    def _store_mediainfos_by_file(self, filepath, mediainfo):
        self._mediainfos_by_file[filepath] = mediainfo

    @property
    def mediainfos_by_file(self):
        """
        Map source video file paths to ``mediainfo`` outputs gathered so far

        .. note:: For DVDs (directories with a VIDEO_TS subdirectory), one
                  mediainfo is made for an ``.IFO`` file and a second mediainfo
                  is made for a ``.VOB`` file.
        """
        return {
            filepath: mediainfo
            for filepath, mediainfo in self._mediainfos_by_file.items()
        }



"""
Concrete :class:`~.base.TrackerJobsBase` subclass for MTV
"""

import functools
import re

from ... import __homepage__, __project_name__, errors, jobs, utils
from ...utils.release import ReleaseType
from ..base import TrackerJobsBase

import logging  # isort:skip
_log = logging.getLogger(__name__)


class MtvTrackerJobs(TrackerJobsBase):

    @functools.cached_property
    def jobs_before_upload(self):
        return (
            # Background jobs
            self.create_torrent_job,
            self.mediainfo_job,
            self.screenshots_job,
            self.upload_screenshots_job,

            # Interactive jobs
            self.category_job,
            self.imdb_job,
            self.scene_check_job,
            self.title_job,
            self.description_job,
        )

    @property
    def isolated_jobs(self):
        if self.options.get('only_description', False):
            return self.get_job_and_dependencies(
                self.description_job,
                # `screenshots_job` is needed by `upload_screenshots_job`, but
                # `upload_screenshots_job` is a `QueueJobBase`, which doesn't
                # know anything about the job it gets input from and therefore
                # can't tells us that it needs `screenshots_job`.
                self.screenshots_job,
            )
        elif self.options.get('only_title', False):
            return self.get_job_and_dependencies(
                self.title_job,
            )
        else:
            # Activate all jobs
            return ()

    @functools.cached_property
    def category_job(self):
        return jobs.dialog.ChoiceJob(
            name=self.get_job_name('category'),
            label='Category',
            precondition=self.make_precondition('category_job'),
            autodetect=self.autodetect_category,
            choices=[
                (c['label'], c['value'])
                for c in self._categories
            ],
            callbacks={
                'finished': self.update_imdb_query,
            },
            **self.common_job_args(),
        )

    def update_imdb_query(self, category_job_):
        """
        Set :attr:`~.webdbs.common.Query.type` on
        :attr:`~.TrackerJobsBase.imdb_job` to :attr:`chosen_release_type`
        """
        old_type = self.imdb_job.query.type
        new_type = self.chosen_release_type
        if new_type is not None:
            _log.debug('Updating IMDb query from %s to %s', old_type, new_type)
            self.imdb_job.query.type = new_type

    _categories = (
        {'label': 'HD Season', 'value': '5', 'type': ReleaseType.season},
        {'label': 'HD Episode', 'value': '3', 'type': ReleaseType.episode},
        {'label': 'HD Movie', 'value': '1', 'type': ReleaseType.movie},
        {'label': 'SD Season', 'value': '6', 'type': ReleaseType.season},
        {'label': 'SD Episode', 'value': '4', 'type': ReleaseType.episode},
        {'label': 'SD Movie', 'value': '2', 'type': ReleaseType.movie},
    )

    _category_value_type_map = {
        c['value']: c['type']
        for c in _categories
    }

    def autodetect_category(self, _):
        # "HD" or "SD"
        if (
            utils.video.height(self.content_path) >= 720
            or utils.video.width(self.content_path) >= 1280
        ):
            resolution = 'HD'
        else:
            resolution = 'SD'

        # "Movie", "Episode" or "Season"
        if self.release_name.type is ReleaseType.movie:
            typ = 'Movie'
        elif self.release_name.type is ReleaseType.season:
            typ = 'Season'
        elif self.release_name.type is ReleaseType.episode:
            typ = 'Episode'
        else:
            raise RuntimeError(f'Unsupported type: {self.release_name.type}')

        category = f'{resolution} {typ}'
        _log.debug('Autodetected category: %r', category)
        return category

    @property
    def chosen_release_type(self):
        """
        :class:`~.types.ReleaseType` enum derived from :attr:`category_job` or
        `None` if :attr:`category_job` is not finished yet
        """
        if self.category_job.is_finished:
            choice = self.get_job_attribute(self.category_job, 'choice')
            return self._category_value_type_map.get(choice)

    release_name_separator = '.'

    release_name_translation = {
        'edition': {
            re.compile(r"^Director's Cut$"): r'DC',
        },
        'group': {
            re.compile(r'^NOGROUP$'): r'NOGRP',
        },
    }

    @functools.cached_property
    def title_job(self):
        """
        :class:`~.jobs.dialog.TextFieldJob` instance with text set to the
        release title

        Unlike :attr:`~.TrackerJobsBase.release_name_job`, this uses the
        original scene release name for movie and episode scene releases.
        """
        return jobs.dialog.TextFieldJob(
            name=self.get_job_name('title'),
            label='Title',
            precondition=self.make_precondition('title_job'),
            prejobs=(
                self.category_job,
                self.scene_check_job,
                self.imdb_job,
            ),
            text=self.generate_title,
            validator=self.validate_title,
            **self.common_job_args(),
        )

    async def generate_title(self):
        assert self.scene_check_job.is_finished
        assert self.imdb_job.is_finished

        try:
            if self.chosen_release_type in (ReleaseType.movie, ReleaseType.episode):
                if self.scene_check_job.is_scene_release:
                    # Use the original scene release name instead of a
                    # standardized release name. We rely on scene_check_job
                    # making sure that the file/directory name is the correct
                    # release name.
                    search_results = await utils.predbs.MultiPredbApi().search(self.content_path)
                    assert len(search_results) >= 1, search_results
                    return search_results[0]

            # We're dealing with a non-scene release or season pack.
            # We may generate a title.
            await self.release_name.fetch_info(webdb=self.imdb, webdb_id=self.imdb_id)

        except errors.RequestError as e:
            _log.debug('Fetching title failed: %r', e)

        return str(self.release_name)

    def validate_title(self, text):
        if not text.strip():
            raise ValueError('Title must not be empty')

        match = re.search(rf'(?:{utils.release.DELIM}|$)(UNKNOWN_([A-Z_]+))(?:{utils.release.DELIM}|$)', text)
        if match:
            placeholder = match.group(1)
            name = match.group(2).lower().replace('_', ' ')
            raise ValueError(f'Replace "{placeholder}" with the correct {name}')

    image_host_config = {
        'common': {'thumb_width': 300},
    }

    @functools.cached_property
    def description_job(self):
        return jobs.dialog.TextFieldJob(
            name=self.get_job_name('description'),
            label='Description',
            precondition=self.make_precondition('description_job'),
            prejobs=(
                self.mediainfo_job,
                self.upload_screenshots_job,
            ),
            text=self.generate_description,
            finish_on_success=True,
            read_only=True,
            hidden=True,
            # Don't cache job output because the number of screenshots can be
            # changed by the user between runs.
            **self.common_job_args(ignore_cache=True),
        )

    async def generate_description(self):
        sections = []
        section = []
        for video_path, info in self.mediainfos_and_screenshots.items():
            if info['mediainfo']:
                section.append(
                    f'[mediainfo]{info["mediainfo"]}[/mediainfo]\n'
                )

            if info['screenshot_urls']:
                section.append((
                    '[center]'
                    + self.make_screenshots_grid(
                        screenshots=info['screenshot_urls'],
                        columns=2,
                        horizontal_spacer='    ',
                        vertical_spacer='\n\n',
                    )
                    + '[/center]'
                ))
                sections.append(''.join(section))
                section.clear()

        description = '\n[hr]\n'.join(sections)
        promotion = (
            '[align=right][size=1]'
            f'Shared with [url={__homepage__}]{__project_name__}[/url]'
            '[/size][/align]'
        )
        return (
            description
            + '\n'
            + promotion
        )

    @property
    def post_data_autofill(self):
        return {
            'submit': 'true',
            'MAX_FILE_SIZE': '2097152',
            'fillonly': 'auto fill',
            'category': '0',
            'Resolution': '0',
            'source': '12',
            'origin': '6',
            'title': '',
            'genre_tags': '---',
            'taglist': '',
            'autocomplete_toggle': 'on',
            'image': '',
            'desc': '',
            'fontfont': '-1',
            'fontsize': '-1',
            'groupDesc': '',
            'anonymous': '0',
        }

    @property
    def post_data_upload(self):
        return {
            'submit': 'true',
            'category': self.get_job_attribute(self.category_job, 'choice'),
            'Resolution': '0',
            'source': '12',
            'origin': '6',
            'title': self.get_job_output(self.title_job, slice=0),
            'genre_tags': '---',
            'autocomplete_toggle': 'on',
            'image': '',
            'desc': self.get_job_output(self.description_job, slice=0),
            'fontfont': '-1',
            'fontsize': '-1',
            'groupDesc': '',
            'anonymous': '1' if self.options['anonymous'] else '0',
            'ignoredupes': '1' if self.options['ignore_dupes'] else None,
            'imdbID': self.get_job_output(self.imdb_job, slice=0),
            # 'tmdbID': ...,
            # 'thetvdbID': ...,
            # 'tvmazeID': ...,
        }

    @property
    def torrent_filepath(self):
        return self.get_job_output(self.create_torrent_job, slice=0)



from types import SimpleNamespace

import pytest
from bs4 import BeautifulSoup

from upsies.utils import html


class Soup(BeautifulSoup):
    def __init__(self, html):
        super().__init__(html, features='html.parser')


@pytest.mark.parametrize(
    argnames='string, exp_prettified_html',
    argvalues=(
        ('<html>foo</html>', '<html>\n foo\n</html>'),
        ('<html>foo</html', '<html>\n foo&lt;/html\n</html>'),
        (Soup('<html>foo</html>'), '<html>\n foo\n</html>'),
        (Soup('<html>foo<b>B<i>A</i>R</b></html>').html.b, '<b>\n B\n <i>\n  A\n </i>\n R\n</b>'),
    ),
    ids=lambda v: repr(v),
)
def test_parse(string, exp_prettified_html):
    assert html.parse(string).prettify().strip() == exp_prettified_html.strip()


def test_dump_writes_string(tmp_path):
    html.dump('<html>foo</html>', tmp_path / 'foo.html')
    assert (tmp_path / 'foo.html').read_text().strip() == '<html>\n foo\n</html>'

def test_dump_writes_BeautifulSoup_instance(tmp_path):
    from bs4 import BeautifulSoup
    doc = BeautifulSoup('<html>foo</html>', features='html.parser')
    html.dump(doc, tmp_path / 'foo.html')
    assert (tmp_path / 'foo.html').read_text().strip() == '<html>\n foo\n</html>'


@pytest.mark.parametrize(
    argnames='html_string, exp_text',
    argvalues=(
        ('<html>foo</html>', 'foo'),
        ('\n<html>  f o   o\n\n bar</html>\n\n', 'f o o\nbar'),
        ('<textarea>Check <b>this</b>   out<i>! </i></textarea>\n\n', 'Check this out!'),
        ('<textarea>Multiple\n<br />lines<br>with\n<br/><br>different<br />\n<br />newlines\n</textarea>\n\n',
         'Multiple\nlines\nwith\ndifferent\nnewlines'),
    ),

)
def test_as_text(html_string, exp_text):
    assert html.as_text(html_string) == exp_text


class Soup(SimpleNamespace):
    pass

@pytest.mark.parametrize(
    argnames='doc, attributes, exp_return_value',
    argvalues=(
        # All tags are found
        (
            Soup(table=Soup(tr=Soup(td='hello'))),
            ('table', 'tr', 'td'),
            'hello',
        ),
        (
            Soup(table=Soup(tr=Soup(td='hello'))),
            ('table', 'tr'),
            Soup(td='hello'),
        ),
        (
            Soup(table=Soup(tr=Soup(td='hello'))),
            ('table',),
            Soup(tr=Soup(td='hello')),
        ),
        (
            Soup(table=Soup(tr=Soup(td='hello'))),
            (),
            Soup(table=Soup(tr=Soup(td='hello'))),
        ),
        # Tag is None
        (
            Soup(table=Soup(tr=Soup(td=None))),
            ('table', 'tr', 'td'),
            None,
        ),
        (
            Soup(table=Soup(tr=Soup(td=None))),
            ('table', 'tr', 'td', 'div'),
            None,
        ),
        # Tag is not found
        (
            Soup(table=Soup(tr=Soup(td='hello'))),
            ('table', 'tr', 'TD'),
            None,
        ),
        (
            Soup(table=Soup(tr=Soup(td='hello'))),
            ('table', 'TR', 'td'),
            None,
        ),
        (
            Soup(table=Soup(tr=Soup(td='hello'))),
            ('TABLE', 'tr', 'td'),
            None,
        ),
    ),
    ids=lambda v: repr(v),
)
def test_get(doc, attributes, exp_return_value):
    return_value = html.get(doc, *attributes)
    assert return_value == exp_return_value


def test_purge_javascript():
    string = (
        '<html>\n'
        'a\n'
        '<script>foo</script>\n'
        'b<script with="attribute">bar</script>\n'
        'c\n'
        '<script>\n'
        '    baz\n'
        '</script>\n'
        '</html>\n'
    )
    assert html.purge_javascript(string) == (
        '<html>\n'
        'a\n'
        '\n'
        'b\n'
        'c\n'
        '\n'
        '</html>\n'
    )



import io
from unittest.mock import AsyncMock, Mock, call

import pytest

from upsies import errors
from upsies.trackers.bhd import BhdTracker, BhdTrackerConfig, BhdTrackerJobs
from upsies.utils.http import Response


def test_name_attribute():
    assert BhdTracker.name == 'bhd'


def test_label_attribute():
    assert BhdTracker.label == 'BHD'


def test_TrackerConfig_attribute():
    assert BhdTracker.TrackerConfig is BhdTrackerConfig


def test_TrackerJobs_attribute():
    assert BhdTracker.TrackerJobs is BhdTrackerJobs


@pytest.mark.asyncio
async def test_login(mocker):
    tracker = BhdTracker()
    await tracker._login()


@pytest.mark.asyncio
async def test_logout(mocker):
    tracker = BhdTracker()
    await tracker._logout()


@pytest.mark.parametrize(
    argnames='announce_url',
    argvalues=(
        'http://bhd.tracker.local:1234/announce',
        'http://bhd.tracker.local:1234/announce/',
    ),
)
@pytest.mark.asyncio
async def test_get_announce_url(announce_url):
    tracker = BhdTracker(options={
        'announce_url': announce_url,
        'announce_passkey': 'd34db33f',
    })
    url = await tracker.get_announce_url()
    assert url == announce_url.rstrip('/') + '/d34db33f'


@pytest.mark.parametrize(
    argnames='upload_url',
    argvalues=(
        'http://bhd/upload',
        'http://bhd/upload/',
    ),
)
def test_get_upload_url(upload_url):
    tracker = BhdTracker(options={
        'upload_url': upload_url,
        'apikey': 'd34db33f',
    })
    url = tracker.get_upload_url()
    assert url == upload_url.rstrip('/') + '/d34db33f'

def test_get_upload_url_without_apikey():
    tracker = BhdTracker(options={
        'upload_url': 'http://localhost/upload',
        'apikey': '',
    })
    with pytest.raises(errors.RequestError, match=rf'^trackers.{tracker.name}.apikey is not set$'):
        tracker.get_upload_url()


@pytest.mark.asyncio
async def test_upload_gets_invalid_json(mocker):
    tracker = BhdTracker(options={'upload_url': 'http://bhd.local/upload', 'apikey': '1337'})
    tracker_jobs_mock = Mock(
        post_data={'foo': 'bar'},
        torrent_filepath='path/to/content.torrent',
        mediainfo_filehandle=io.StringIO('mediainfo mock'),
    )
    http_mock = mocker.patch('upsies.utils.http', Mock(
        post=AsyncMock(return_value=Response(
            text='{choke on this',
            bytes=b'irrelevant',
        )),
    ))
    with pytest.raises(errors.RequestError, match=r'^Malformed JSON: \{choke on this'):
        await tracker.upload(tracker_jobs_mock)
    assert http_mock.post.call_args_list == [call(
        url='http://bhd.local/upload/1337',
        cache=False,
        user_agent=True,
        data=tracker_jobs_mock.post_data,
        files={
            'file': {
                'file': tracker_jobs_mock.torrent_filepath,
                'mimetype': 'application/octet-stream',
            },
            'mediainfo': {
                'file': tracker_jobs_mock.mediainfo_filehandle,
                'filename': 'mediainfo',
                'mimetype': 'application/octet-stream',
            },
        },
    )]

@pytest.mark.asyncio
async def test_upload_gets_error_status_code(mocker):
    tracker = BhdTracker(options={'upload_url': 'http://bhd.local/upload', 'apikey': '1337'})
    tracker_jobs_mock = Mock(
        post_data={'foo': 'bar'},
        torrent_filepath='path/to/content.torrent',
        mediainfo_filehandle=io.StringIO('mediainfo mock'),
    )
    http_mock = mocker.patch('upsies.utils.http', Mock(
        post=AsyncMock(return_value=Response(
            text='''
            {
                "status_code": 0,
                "success": false,
                "status_message": "This is the error message"
            }
            ''',
            bytes=b'irrelevant',
        )),
    ))
    with pytest.raises(errors.RequestError, match=r'^Upload failed: This is the error message$'):
        await tracker.upload(tracker_jobs_mock)
    assert http_mock.post.call_args_list == [call(
        url='http://bhd.local/upload/1337',
        cache=False,
        user_agent=True,
        data=tracker_jobs_mock.post_data,
        files={
            'file': {
                'file': tracker_jobs_mock.torrent_filepath,
                'mimetype': 'application/octet-stream',
            },
            'mediainfo': {
                'file': tracker_jobs_mock.mediainfo_filehandle,
                'filename': 'mediainfo',
                'mimetype': 'application/octet-stream',
            },
        },
    )]

@pytest.mark.asyncio
async def test_upload_gets_unexpected_status_code(mocker):
    tracker = BhdTracker(options={'upload_url': 'http://bhd.local/upload', 'apikey': '1337'})
    tracker_jobs_mock = Mock(
        post_data={'foo': 'bar'},
        torrent_filepath='path/to/content.torrent',
        mediainfo_filehandle=io.StringIO('mediainfo mock'),
    )
    http_mock = mocker.patch('upsies.utils.http', Mock(
        post=AsyncMock(return_value=Response(
            text='{"status_code": 123.5}',
            bytes=b'irrelevant',
        )),
    ))
    with pytest.raises(RuntimeError, match=r"^Unexpected response: '\{\"status_code\": 123.5\}'$"):
        await tracker.upload(tracker_jobs_mock)
    assert http_mock.post.call_args_list == [call(
        url='http://bhd.local/upload/1337',
        cache=False,
        user_agent=True,
        data=tracker_jobs_mock.post_data,
        files={
            'file': {
                'file': tracker_jobs_mock.torrent_filepath,
                'mimetype': 'application/octet-stream',
            },
            'mediainfo': {
                'file': tracker_jobs_mock.mediainfo_filehandle,
                'filename': 'mediainfo',
                'mimetype': 'application/octet-stream',
            },
        },
    )]

@pytest.mark.asyncio
async def test_upload_gets_unexpected_json(mocker):
    tracker = BhdTracker(options={'upload_url': 'http://bhd.local/upload', 'apikey': '1337'})
    tracker_jobs_mock = Mock(
        post_data={'foo': 'bar'},
        torrent_filepath='path/to/content.torrent',
        mediainfo_filehandle=io.StringIO('mediainfo mock'),
    )
    http_mock = mocker.patch('upsies.utils.http', Mock(
        post=AsyncMock(return_value=Response(
            text='{"hey": "you"}',
            bytes=b'irrelevant',
        )),
    ))
    with pytest.raises(RuntimeError, match=r"^Unexpected response: '\{\"hey\": \"you\"\}'$"):
        await tracker.upload(tracker_jobs_mock)
    assert http_mock.post.call_args_list == [call(
        url='http://bhd.local/upload/1337',
        cache=False,
        user_agent=True,
        data=tracker_jobs_mock.post_data,
        files={
            'file': {
                'file': tracker_jobs_mock.torrent_filepath,
                'mimetype': 'application/octet-stream',
            },
            'mediainfo': {
                'file': tracker_jobs_mock.mediainfo_filehandle,
                'filename': 'mediainfo',
                'mimetype': 'application/octet-stream',
            },
        },
    )]

@pytest.mark.asyncio
async def test_upload_succeeds(mocker):
    tracker = BhdTracker(options={'upload_url': 'http://bhd.local/upload', 'apikey': '1337'})
    tracker_jobs_mock = Mock(
        post_data={
            'foo': 'asdf',
            'bar': '',
            'baz': 0,
            'quux': '0',
            'quuz': None,
        },
        torrent_filepath='path/to/content.torrent',
        mediainfo_filehandle=io.StringIO('mediainfo mock'),
    )
    http_mock = mocker.patch('upsies.utils.http', Mock(
        post=AsyncMock(return_value=Response(
            text='''
            {
                "status_code": 2,
                "success": true,
                "status_message": "http://bhd.local/torrent/download/release_name.123456.d34db33f"
            }
            ''',
            bytes=b'irrelevant',
        )),
    ))
    torrent_page_url = await tracker.upload(tracker_jobs_mock)
    assert torrent_page_url == 'http://bhd.local/torrents/release_name.123456'
    assert http_mock.post.call_args_list == [call(
        url='http://bhd.local/upload/1337',
        cache=False,
        user_agent=True,
        data={
            'foo': 'asdf',
            'baz': '0',
            'quux': '0',
        },
        files={
            'file': {
                'file': tracker_jobs_mock.torrent_filepath,
                'mimetype': 'application/octet-stream',
            },
            'mediainfo': {
                'file': tracker_jobs_mock.mediainfo_filehandle,
                'filename': 'mediainfo',
                'mimetype': 'application/octet-stream',
            },
        },
    )]

@pytest.mark.asyncio
async def test_upload_draft_succeeds(mocker):
    tracker = BhdTracker(options={'upload_url': 'http://bhd.local/upload', 'apikey': '1337'})
    warning_cb = Mock()
    tracker.signal.register('warning', warning_cb)
    tracker_jobs_mock = Mock(
        post_data={'foo': 'bar'},
        torrent_filepath='path/to/content.torrent',
        mediainfo_filehandle=io.StringIO('mediainfo mock'),
    )
    http_mock = mocker.patch('upsies.utils.http', Mock(
        post=AsyncMock(return_value=Response(
            text='''
            {
                "status_code": 1,
                "success": true,
                "status_message": "Draft uploaded"
            }
            ''',
            bytes=b'irrelevant',
        )),
    ))
    torrent_filepath = await tracker.upload(tracker_jobs_mock)
    assert warning_cb.call_args_list == [
        call('Draft uploaded'),
        call('You have to activate your upload manually '
             'on the website when you are ready to seed.')
    ]
    assert torrent_filepath == tracker_jobs_mock.torrent_filepath
    assert http_mock.post.call_args_list == [call(
        url='http://bhd.local/upload/1337',
        cache=False,
        user_agent=True,
        data=tracker_jobs_mock.post_data,
        files={
            'file': {
                'file': tracker_jobs_mock.torrent_filepath,
                'mimetype': 'application/octet-stream',
            },
            'mediainfo': {
                'file': tracker_jobs_mock.mediainfo_filehandle,
                'filename': 'mediainfo',
                'mimetype': 'application/octet-stream',
            },
        },
    )]


@pytest.mark.parametrize(
    argnames='bytes, exp_piece_size',
    argvalues=(
        # 1 MiB
        (1, 1 * (1024 ** 2)),
        (8 * (1024 ** 3) - 1, 1 * (1024 ** 2)),
        # 2 MiB
        (8 * (1024 ** 3), 2 * (1024 ** 2)),
        (16 * (1024 ** 3) - 1, 2 * (1024 ** 2)),
        # 4 MiB
        (16 * (1024 ** 3), 4 * (1024 ** 2)),
        (72 * (1024 ** 3) - 1, 4 * (1024 ** 2)),
        # 8 MiB
        (72 * (1024 ** 3), 8 * (1024 ** 2)),
        (float('inf'), 8 * (1024 ** 2)),
    ),
)
def test_calculate_piece_size(bytes, exp_piece_size):
    piece_size = BhdTracker.calculate_piece_size(bytes)
    assert piece_size == exp_piece_size



import asyncio
from unittest.mock import AsyncMock, Mock, call

import pytest

from upsies import errors
from upsies.jobs.custom import CustomJob


@pytest.fixture
def make_CustomJob(tmp_path):
    def make_CustomJob(**kwargs):
        return CustomJob(home_directory=tmp_path, cache_directory=tmp_path, **kwargs)
    return make_CustomJob


def test_name(make_CustomJob):
    job = make_CustomJob(name='foo', label='Foo', worker=AsyncMock())
    assert job.name == 'foo'


def test_label(make_CustomJob):
    job = make_CustomJob(name='foo', label='Foo', worker=AsyncMock())
    assert job.label == 'Foo'


def test_initialize(make_CustomJob):
    worker = AsyncMock()
    job = make_CustomJob(name='foo', label='Foo', worker=worker)
    assert job._worker is worker
    assert job._task is None


@pytest.mark.asyncio
async def test_execute(mocker, make_CustomJob):
    mocker.patch('upsies.jobs.custom.CustomJob.attach_task')
    mocker.patch('upsies.jobs.custom.CustomJob._catch_errors', Mock())
    worker = Mock()
    job = make_CustomJob(name='foo', label='Foo', worker=worker)
    job.execute()
    assert job._task is job.attach_task.return_value
    assert job.attach_task.call_args_list == [call(
        job._catch_errors.return_value,
        callback=job._handle_worker_done,
    )]
    assert job._catch_errors.call_args_list == [call(worker.return_value)]
    assert worker.call_args_list == [call(job)]


@pytest.mark.asyncio
async def test_CancelledError_from_worker_is_ignored(make_CustomJob):
    worker = AsyncMock(side_effect=asyncio.CancelledError())
    job = make_CustomJob(name='foo', label='Foo', worker=worker)
    job.execute()
    # Does not raise CancelledError
    await job._task

@pytest.mark.asyncio
async def test_expected_exceptions_from_worker_are_caught(mocker, make_CustomJob):
    worker = AsyncMock(side_effect=errors.UpsiesError('ouch'))
    job = make_CustomJob(name='foo', label='Foo', worker=worker, catch=[errors.UpsiesError])
    job.execute()
    await job._task
    assert job.errors == (errors.UpsiesError('ouch'),)

@pytest.mark.asyncio
async def test_unexpected_exceptions_from_worker_are_raised(mocker, make_CustomJob):
    worker = AsyncMock(side_effect=errors.UpsiesError('ouch'))
    job = make_CustomJob(name='foo', label='Foo', worker=worker, catch=[TypeError])
    job.execute()
    mocker.patch.object(job, 'exception')
    with pytest.raises(errors.UpsiesError, match=r'^ouch$'):
        await job.await_tasks()
    assert job.exception.call_args_list == [call(errors.UpsiesError('ouch'))]


@pytest.mark.parametrize(
    argnames='return_value, exp_output',
    argvalues=(
        ('foo', ('foo',)),
        ('', ('',)),
        (0, ('0',)),
        (None, ()),
    ),
    ids=lambda v: str(v),
)
@pytest.mark.asyncio
async def test_return_value_from_worker_is_output(return_value, exp_output, mocker, make_CustomJob):
    worker = AsyncMock(return_value=return_value)
    job = make_CustomJob(name='foo', label='Foo', worker=worker)
    job.execute()
    await job.wait()
    assert job.output == exp_output


@pytest.mark.asyncio
async def test_wait_before_job_was_executed(mocker, make_CustomJob):
    job = make_CustomJob(name='foo', label='Foo', worker=AsyncMock())
    assert job._task is None
    job.finish()
    await job.wait()
    assert job._task is None
    assert job.is_finished

@pytest.mark.asyncio
async def test_wait_after_job_was_executed(mocker, make_CustomJob):
    job = make_CustomJob(name='foo', label='Foo', worker=lambda job: asyncio.sleep(10))
    job.execute()
    # Allow worker to start
    for _ in range(5):
        await asyncio.sleep(0)
    assert not job._task.done()
    job.finish()
    await job.wait()
    assert job._task.done()
    assert job.is_finished


@pytest.mark.asyncio
async def test_finish_before_job_was_executed(mocker, make_CustomJob):
    job = make_CustomJob(name='foo', label='Foo', worker=AsyncMock())
    refresh_job_list_cb = Mock()
    job.signal.register('refresh_job_list', refresh_job_list_cb)
    assert job._task is None
    job.finish()
    assert job._task is None
    assert job.is_finished
    assert refresh_job_list_cb.call_args_list == []


@pytest.mark.asyncio
async def test_finish_after_job_was_executed(mocker, make_CustomJob):
    worker = asyncio.ensure_future(asyncio.sleep(100))
    job = make_CustomJob(name='foo', label='Foo', worker=lambda job: worker)
    refresh_job_list_cb = Mock()
    job.signal.register('refresh_job_list', refresh_job_list_cb)
    job.execute()
    # Allow worker to start
    for _ in range(5):
        await asyncio.sleep(0)
    assert not worker.cancelled()
    job.finish()
    await job.wait()
    assert worker.cancelled()
    assert job.is_finished
    assert refresh_job_list_cb.call_args_list == [call()]



import re
import types
from unittest.mock import AsyncMock, Mock, call

import aiobtclientapi
import aiobtclientrpc
import pytest

from upsies import errors, utils
from upsies.utils import btclient


def test_client_names(mocker):
    client_names_mock = mocker.patch('aiobtclientapi.client_names')
    return_value = btclient.client_names()
    assert return_value is client_names_mock.return_value
    client_names_mock.call_args_list == [call()]


@pytest.mark.parametrize(
    argnames='client_name, exp_config',
    argvalues=(
        ('foo', None),
        (
            'deluge', {
                'url': aiobtclientapi.api_class('deluge').URL.default,
                'username': '',
                'password': '',
                'check_after_add': utils.types.Bool('no'),
            },
        ),
        (
            'qbittorrent', {
                'url': aiobtclientapi.api_class('qbittorrent').URL.default,
                'username': '',
                'password': '',
                'check_after_add': utils.types.Bool('no'),
                'category': '',
            },
        ),
        (
            'rtorrent', {
                'url': aiobtclientapi.api_class('rtorrent').URL.default,
                'username': '',
                'password': '',
                'check_after_add': utils.types.Bool('no'),
            },
        ),
        (
            'transmission', {
                'url': aiobtclientapi.api_class('transmission').URL.default,
                'username': '',
                'password': '',
                'check_after_add': utils.types.Bool('no'),
            },
        ),
    ),
    ids=lambda v: repr(v),
)
def test_client_defaults(client_name, exp_config, mocker):
    return_value = btclient.client_defaults(client_name)
    assert return_value == exp_config


@pytest.mark.parametrize(
    argnames='client_name, kwargs, exp_calls',
    argvalues=(
        (
            'foo',
            {
                'url': 'http://foo',
                'username': 'bar',
                'password': 'baz',
                'download_path': 'some/path',
                'check_after_add': 'maybe',
            },
            [
                call.api(
                    name='foo',
                    url='http://foo',
                    username='bar',
                    password='baz',
                ),
            ],
        ),
        (
            'foo',
            {
                'url': 'http://foo',
                'username': 'bar',
                'password': 'baz',
                'download_path': 'some/path',
                'check_after_add': 'maybe',
                'category': 'my_category',
            },
            [
                call.api(
                    name='foo',
                    url='http://foo',
                    username='bar',
                    password='baz',
                ),
            ],
        ),
    ),
    ids=lambda v: repr(v),
)
def test_BtClient_api(client_name, kwargs, exp_calls, mocker):
    mocks = Mock()
    mocks.attach_mock(mocker.patch('aiobtclientapi.api', return_value='mock api'), 'api')
    client = btclient.BtClient(client_name, **kwargs)
    assert client._api == 'mock api'
    assert client._download_path == kwargs['download_path']
    assert client._check_after_add == kwargs['check_after_add']
    assert client._category == kwargs.get('category', '')
    assert mocks.mock_calls == exp_calls


def test_BtClient_name(mocker):
    api = Mock()
    api.configure_mock(name='myclient')
    mocker.patch('aiobtclientapi.api', return_value=api)

    client = btclient.BtClient(
        name='foo',
        url='http://foo',
        username='bar',
        password='baz',
        download_path='some/path',
        check_after_add='maybe',
    )

    assert client.name == 'myclient'

    # Property is immutable
    with pytest.raises(AttributeError):
        client.name = 'notmyclient'


def test_BtClient_label(mocker):
    api = Mock()
    api.configure_mock(label='MyClient')
    mocker.patch('aiobtclientapi.api', return_value=api)

    client = btclient.BtClient(
        name='irrelevant',
        url='http://foo',
        username='bar',
        password='baz',
        download_path='some/path',
        check_after_add='maybe',
    )

    assert client.label == 'MyClient'

    # Property is immutable
    with pytest.raises(AttributeError):
        client.label = 'NoyMyClient'


@pytest.mark.parametrize('category', ('', 'my_category'))
@pytest.mark.parametrize(
    argnames=(
        'torrent, download_path, check_after_add, '
        'response_errors, response_added, response_already_added, '
        'exp_result, exp_calls'
    ),
    argvalues=(
        pytest.param(
            'my.torrent', 'download/path', 'maybe verify',
            [Exception('whatever')], [], [],
            errors.TorrentAddError('whatever'),
            [
                call.add(
                    'my.torrent',
                    location='download/path',
                    verify='maybe verify',
                ),
            ],
            id='Adding torrent fails',
        ),
        pytest.param(
            'my.torrent', 'download/path', 'maybe verify',
            [], ['d34db33f'], [],
            'd34db33f',
            [
                call.add(
                    'my.torrent',
                    location='download/path',
                    verify='maybe verify',
                ),
            ],
            id='Adding torrent succeeds',
        ),
    ),
    ids=lambda v: repr(v),
)
@pytest.mark.asyncio
async def test_BtClient_add_torrent(torrent, download_path, check_after_add, category,
                                    response_errors, response_added, response_already_added,
                                    exp_result, exp_calls,
                                    mocker):
    mocker.patch('aiobtclientapi.api', return_value=Mock(
        add=AsyncMock(
            return_value=Mock(
                errors=response_errors,
                added=response_added,
                already_added=response_already_added,
            ),
        ),
    ))

    client = btclient.BtClient(
        name='irrelevant',
        url='http://foo',
        username='bar',
        password='baz',
        download_path=download_path,
        check_after_add=check_after_add,
        category=category,
    )

    mocks = Mock()
    mocks.attach_mock(client._api.add, 'add')
    mocks.attach_mock(mocker.patch.object(client, '_set_category'), '_set_category')

    if isinstance(exp_result, Exception):
        with pytest.raises(type(exp_result), match=rf'^{re.escape(str(exp_result))}$'):
            await client.add_torrent(torrent)
    else:
        return_value = await client.add_torrent(torrent)
        assert return_value == exp_result

        if category:
            exp_calls.append(call._set_category(exp_result, category))

    assert mocks.mock_calls == exp_calls


@pytest.mark.parametrize(
    argnames='infohash, category, api, exp_exception, exp_calls',
    argvalues=(
        pytest.param(
            'd34db33f',
            'my_category',
            types.SimpleNamespace(
                name='notqbittorrent',
                label='NotqBittorrent',
                call=AsyncMock(),
            ),
            RuntimeError('Categories are not supported for NotqBittorrent'),
            [],
            id='Client is not qBittorrent',
        ),
        pytest.param(
            'd34db33f',
            'my_category',
            types.SimpleNamespace(
                name='qbittorrent',
                label='qBittorrent',
                call=AsyncMock(side_effect=aiobtclientrpc.RPCError('Incorrect category name')),
            ),
            errors.TorrentAddError('Unknown category: my_category'),
            [],
            id='Setting unknown category',
        ),
        pytest.param(
            'd34db33f',
            'my_category',
            types.SimpleNamespace(
                name='qbittorrent',
                label='qBittorrent',
                call=AsyncMock(side_effect=aiobtclientrpc.RPCError('WAT')),
            ),
            aiobtclientrpc.RPCError('WAT'),
            [],
            id='Unexpected exception',
        ),
    ),
    ids=lambda v: repr(v),
)
@pytest.mark.asyncio
async def test_BtClient__set_category(infohash, category, api, exp_exception, exp_calls, mocker):
    mocker.patch('aiobtclientapi.api', return_value=api)

    client = btclient.BtClient(
        name='irrelevant',
        url='http://foo',
        username='bar',
        password='baz',
        download_path='download/path',
        check_after_add='maybe',
        category=category,
    )

    mocks = Mock()
    mocks.attach_mock(client._api, 'api')

    if isinstance(exp_exception, Exception):
        with pytest.raises(type(exp_exception), match=rf'^{re.escape(str(exp_exception))}$'):
            await client._set_category(infohash, category)
    else:
        return_value = await client._set_category(infohash, category)
        assert return_value is None

    assert mocks.mock_calls == exp_calls



import base64
import re

import pytest

from upsies import utils
from upsies.trackers.mtv import MtvTrackerConfig


@pytest.fixture
def tracker_config():
    return MtvTrackerConfig()


def test_MtvTrackerConfig_defaults(tracker_config):
    assert set(tracker_config) == {
        'base_url',
        'username',
        'password',
        'announce_url',
        'source',
        'randomize_infohash',
        'image_host',
        'screenshots',
        'exclude',
        'anonymous',

        # Inherited from TrackerConfigBase
        'add_to',
        'copy_to',
    }


def test_MtvTrackerConfig_defaults_base_url(tracker_config):
    assert tracker_config['base_url'] == base64.b64decode('aHR0cHM6Ly93d3cubW9yZXRoYW50di5tZQ==').decode('ascii')


def test_MtvTrackerConfig_defaults_username(tracker_config):
    assert tracker_config['username'] == ''


def test_MtvTrackerConfig_defaults_password(tracker_config):
    assert tracker_config['password'] == ''


def test_MtvTrackerConfig_defaults_announce_url(tracker_config):
    assert tracker_config['announce_url'] == ''
    assert tracker_config['announce_url'].description == (
        'Your personal announce URL. Automatically fetched from the website if not set.'
    )


def test_MtvTrackerConfig_defaults_source(tracker_config):
    assert tracker_config['source'] == 'MTV'


def test_MtvTrackerConfig_defaults_image_host(tracker_config, assert_config_choice):
    assert_config_choice(
        choice=tracker_config['image_host'],
        value='imgbox',
        options=utils.imghosts.imghost_names(),
    )


def test_MtvTrackerConfig_defaults_screenshots(tracker_config, assert_config_number):
    assert_config_number(
        number=tracker_config['screenshots'],
        value=4,
        min=3,
        max=10,
        description='How many screenshots to make.',
    )


def test_MtvTrackerConfig_defaults_exclude(tracker_config):
    assert tracker_config['exclude'] == (
        re.compile(r'\.(?i:nfo|txt|jpg|jpeg|png|sfv|md5)$'),
        re.compile(r'/(?i:sample|extra|bonus|feature)'),
        re.compile(r'(?i:sample\.[a-z]+)$'),
    )


def test_MtvTrackerConfig_defaults_anonymous(tracker_config):
    assert isinstance(tracker_config['anonymous'], utils.types.Bool)
    assert not tracker_config['anonymous']
    assert tracker_config['anonymous'] == 'no'
    assert tracker_config['anonymous'].description == (
        'Whether your username is displayed on your uploads.'
    )


def test_MtvTrackerConfig_arguments(tracker_config):
    exp_argument_definitions = {
        'submit': {
            ('--screenshots', '--ss'),
            ('--only-description', '--od'),
            ('--only-title', '--ot'),
            ('--ignore-dupes', '--id'),
        },
    }
    assert set(tracker_config.argument_definitions) == set(exp_argument_definitions)
    for command in exp_argument_definitions:
        assert set(tracker_config.argument_definitions[command]) == exp_argument_definitions[command]


def test_MtvTrackerConfig_argument_definitions_submit_screenshots(tracker_config):
    assert tracker_config.argument_definitions['submit'][('--screenshots', '--ss')] == {
        'help': ('How many screenshots to make '
                 f'(min={tracker_config["screenshots"].min}, '
                 f'max={tracker_config["screenshots"].max})'),
        'type': utils.argtypes.number_of_screenshots(
            min=tracker_config['screenshots'].min,
            max=tracker_config['screenshots'].max,
        ),
    }


def test_MtvTrackerConfig_argument_definitions_submit_only_description(tracker_config):
    assert tracker_config.argument_definitions['submit'][('--only-description', '--od')] == {
        'help': 'Only generate description (do not upload anything)',
        'action': 'store_true',
        'group': 'generate-metadata',
    }


def test_MtvTrackerConfig_argument_definitions_submit_only_title(tracker_config):
    assert tracker_config.argument_definitions['submit'][('--only-title', '--ot')] == {
        'help': 'Only generate title (do not upload anything)',
        'action': 'store_true',
        'group': 'generate-metadata',
    }


def test_MtvTrackerConfig_argument_definitions_submit_ignore_dupes(tracker_config):
    assert tracker_config.argument_definitions['submit'][('--ignore-dupes', '--id')] == {
        'help': 'Force submission even if the tracker reports duplicates',
        'action': 'store_true',
    }



"""
Change or show configuration file options
"""

import functools
import textwrap

from .... import constants, defaults, errors, jobs, utils
from .base import CommandBase


def _make_pretty_option_list():
    lines = []
    for option in defaults.option_paths():
        option_type = defaults.option_type(option)
        option_type_name = option_type.__name__.lower()
        file, section, name = option.split('.', maxsplit=2)
        default_value = defaults.defaults[file][section][name]
        text = [f'{option} ({option_type_name})']

        if option_type_name == 'list':
            if default_value:
                text.append(f'  Default: {default_value[0]}')
                for item in default_value[1:]:
                    text.append(f'           {item}')
            else:
                text.append('  Default: <empty>')
        else:
            text.append(f'  Default: {default_value}')

        if option_type_name == 'int':
            option_type_name = 'integer'
        elif option_type_name == 'str':
            option_type_name = 'string'
        elif option_type_name == 'choice':
            text.append(f'  Choices: {", ".join(default_value.options)}')

        # Add description indented
        description = getattr(option_type, 'description', '')
        if description:
            description_lines = []
            for paragraph in description.strip().split('\n'):
                description_lines.extend(
                    textwrap.wrap(
                        text=paragraph,
                        width=72,
                    )
                )
            text.extend('  ' + line for line in description_lines)
        lines.append('\n  '.join(text))

    return '\n\n  '.join(lines)


class set(CommandBase):
    """
    Change or show configuration file options

    Without any arguments, all options are listed with their current values.

    OPTION consists of three segments which are delimited with a period (".").
    The first segment in OPTION is the file name without the extension. The
    second segment is the section name in that file. The third segment is the
    option name.

    List values are given as one argument per list item. If non-list values are
    given as multiple arguments, they are concatenated with single spaces. In
    the INI file, list items are delimited by one line break and one or more
    spaces (e.g. "\\n    ").
    """

    names = ('set',)

    description = 'options:\n  ' + _make_pretty_option_list()

    argument_definitions = {
        'OPTION': {
            'type': utils.argtypes.option,
            'nargs': '?',
            'help': 'Option to change or show',
        },
        'VALUE': {
            'nargs': '*',
            'default': '',  # FIXME: https://bugs.python.org/issue41854
            'help': 'New value for OPTION',
            'group': 'value',
        },
        ('--reset', '-r'): {
            'action': 'store_true',
            'help': 'Reset OPTION to default value',
            'group': 'value',
        },
        ('--fetch-ptpimg-apikey',): {
            'nargs': 2,
            'metavar': ('EMAIL', 'PASSWORD'),
            'help': ('Fetch ptpimg API key and save it in '
                     f'{utils.fs.tildify_path(constants.IMGHOSTS_FILEPATH)}\n'
                     '(EMAIL and PASSWORD are not saved)'),
        },
    }

    @functools.cached_property
    def jobs(self):
        if self.args.fetch_ptpimg_apikey is not None:
            return (self.fetch_ptpimg_apikey_job,)
        else:
            return (
                jobs.set.SetJob(
                    config=self.config,
                    option=self.args.OPTION,
                    # VALUE is a list. The Config class should convert lists to
                    # strings (and vice versa) depending on the default type.
                    value=self.args.VALUE,
                    reset=self.args.reset,
                ),
            )

    @functools.cached_property
    def fetch_ptpimg_apikey_job(self):
        return jobs.custom.CustomJob(
            name='fetch-ptpimg-apikey',
            label='API key',
            worker=self.fetch_ptpimg_apikey,
            catch=(errors.RequestError, errors.ConfigError),
            ignore_cache=self.args.ignore_cache,
        )

    async def fetch_ptpimg_apikey(self, job):
        if len(self.args.fetch_ptpimg_apikey) <= 0:
            job.error('Missing EMAIL and PASSWORD')
        elif len(self.args.fetch_ptpimg_apikey) <= 1:
            job.error('Missing PASSWORD')
        elif len(self.args.fetch_ptpimg_apikey) > 2:
            unknown_args = ' '.join(self.args.fetch_ptpimg_apikey[2:])
            job.error(f'Unrecognized arguments: {unknown_args}')
        else:
            email = self.args.fetch_ptpimg_apikey[0]
            password = self.args.fetch_ptpimg_apikey[1]
            ptpimg = utils.imghosts.imghost('ptpimg')
            apikey = await ptpimg.get_apikey(email, password)
            job.send(apikey)
            self.config['imghosts']['ptpimg']['apikey'] = apikey
            self.config.write('imghosts.ptpimg.apikey')
            job.finish()



"""
Concrete :class:`~.base.TrackerConfigBase` subclass for PTP
"""

import base64

from ...utils import argtypes, configfiles, imghosts, types
from ..base import TrackerConfigBase


class PtpTrackerConfig(TrackerConfigBase):
    defaults = {
        'base_url': base64.b64decode('aHR0cHM6Ly9wYXNzdGhlcG9wY29ybi5tZQ==').decode('ascii'),
        'username': '',
        'password': '',
        'announce_url': configfiles.config_value(
            value='',
            description='Your personal announce URL.',
        ),
        'source': 'PTP',
        'image_host': types.Choice('ptpimg', options=imghosts.imghost_names()),
        'screenshots_from_movie': configfiles.config_value(
            value=types.Integer(3, min=3, max=10),
            description='How many screenshots to make for single-video uploads.',
        ),
        'screenshots_from_episode': configfiles.config_value(
            value=types.Integer(2, min=2, max=10),
            description='How many screenshots to make per video for multi-video uploads.',
        ),
        'exclude': (
            r'\.(?i:nfo|txt|jpg|jpeg|png|sfv|md5)$',
            r'/(?i:sample|extra|bonus|feature)',
            r'(?i:sample\.[a-z]+)$',
        ),
    }

    argument_definitions = {
        'submit': {
            ('--not-main-movie', '--nmm'): {
                'help': 'Upload ONLY contains extras, Rifftrax, Workprints',
                'action': 'store_true',
            },
            ('--personal-rip', '--pr'): {
                'help': 'Tag as your own encode',
                'action': 'store_true',
            },
            ('--screenshots', '--ss'): {
                'help': 'How many screenshots to make per video file',
                'type': argtypes.number_of_screenshots(min=1, max=10),
            },
            ('--upload-token', '--ut'): {
                'help': 'Upload token from staff',
            },
            ('--only-description', '--od'): {
                'help': 'Only generate description (do not upload anything)',
                'action': 'store_true',
                'group': 'generate-metadata',
            },
        },
    }



from unittest.mock import Mock, PropertyMock, call

import pytest

from upsies import utils
from upsies.trackers import nbl


@pytest.fixture
def tracker():
    tracker = Mock()
    tracker.name = 'nbl'
    return tracker


@pytest.fixture
def nbl_tracker_jobs(tracker, tmp_path, mocker):
    content_path = tmp_path / 'Foo S01 1080p BluRay x264-ASDF'

    nbl_tracker_jobs = nbl.NblTrackerJobs(
        content_path=str(content_path),
        tracker=tracker,
        torrent_destination=str(tmp_path / 'destination'),
        common_job_args={
            'home_directory': str(tmp_path / 'home_directory'),
            'ignore_cache': True,
        },
        options=None,
    )

    return nbl_tracker_jobs


def test_jobs_before_upload(nbl_tracker_jobs, tmp_path, mocker):
    create_torrent_job_mock = mocker.patch('upsies.trackers.nbl.NblTrackerJobs.create_torrent_job', Mock())
    mediainfo_job_mock = mocker.patch('upsies.trackers.nbl.NblTrackerJobs.mediainfo_job', Mock())
    tvmaze_job_mock = mocker.patch('upsies.trackers.nbl.NblTrackerJobs.tvmaze_job', Mock())
    category_job_mock = mocker.patch('upsies.trackers.nbl.NblTrackerJobs.category_job', Mock())
    assert nbl_tracker_jobs.jobs_before_upload == (
        create_torrent_job_mock,
        mediainfo_job_mock,
        tvmaze_job_mock,
        category_job_mock,
    )


def test_category_job(nbl_tracker_jobs, mocker):
    ChoiceJob_mock = mocker.patch('upsies.jobs.dialog.ChoiceJob')
    mocker.patch.object(nbl_tracker_jobs, 'get_job_name')
    mocker.patch.object(nbl_tracker_jobs, 'make_precondition')
    mocker.patch.object(type(nbl_tracker_jobs), 'autodetected_category', PropertyMock(return_value='my category'))
    mocker.patch.object(nbl_tracker_jobs, 'common_job_args', return_value={'foo': 'bar'})

    assert nbl_tracker_jobs.category_job is ChoiceJob_mock.return_value
    assert ChoiceJob_mock.call_args_list == [call(
        name=nbl_tracker_jobs.get_job_name.return_value,
        label='Category',
        precondition=nbl_tracker_jobs.make_precondition.return_value,
        choices=(
            ('Season', '3'),
            ('Episode', '1'),
        ),
        autodetected=nbl_tracker_jobs.autodetected_category,
        foo='bar',
    )]
    assert nbl_tracker_jobs.get_job_name.call_args_list == [call('category')]
    assert nbl_tracker_jobs.make_precondition.call_args_list == [call('category_job')]
    assert nbl_tracker_jobs.common_job_args.call_args_list == [call()]


@pytest.mark.parametrize(
    argnames='release_type, exp_category',
    argvalues=(
        (utils.release.ReleaseType.season, 'Season'),
        (utils.release.ReleaseType.episode, 'Episode'),
        (utils.release.ReleaseType.movie, 'Episode'),
    ),
    ids=lambda v: repr(v),
)
def test_autodetected_category(release_type, exp_category, nbl_tracker_jobs, mocker):
    mocker.patch.object(type(nbl_tracker_jobs.release_name), 'type', PropertyMock(return_value=release_type))
    return_value = nbl_tracker_jobs.autodetected_category
    assert return_value == exp_category


def test_post_data(nbl_tracker_jobs, mocker):
    mocker.patch.object(type(nbl_tracker_jobs), 'options', PropertyMock(return_value={
        'apikey': 'mock api key',
    }))
    mocker.patch.object(nbl_tracker_jobs, 'get_job_attribute', side_effect=(
        'mock category',
    ))
    mocker.patch.object(nbl_tracker_jobs, 'get_job_output', side_effect=(
        'mock tvmaze id',
        'mock mediainfo',
    ))
    mocker.patch.object(type(nbl_tracker_jobs), 'category_job', PropertyMock())
    mocker.patch.object(type(nbl_tracker_jobs), 'tvmaze_job', PropertyMock())
    mocker.patch.object(type(nbl_tracker_jobs), 'mediainfo_job', PropertyMock())

    assert nbl_tracker_jobs.post_data == {
        'api_key': 'mock api key',
        'category': 'mock category',
        'tvmazeid': 'mock tvmaze id',
        'mediainfo': 'mock mediainfo',
    }
    assert nbl_tracker_jobs.get_job_attribute.call_args_list == [
        call(nbl_tracker_jobs.category_job, 'choice'),
    ]
    assert nbl_tracker_jobs.get_job_output.call_args_list == [
        call(nbl_tracker_jobs.tvmaze_job, slice=0),
        call(nbl_tracker_jobs.mediainfo_job, slice=0),
    ]


def test_torrent_filepath(nbl_tracker_jobs, mocker):
    mocker.patch.object(type(nbl_tracker_jobs), 'create_torrent_job', PropertyMock(
        return_value='mock create_torrent_job',
    ))
    mocker.patch.object(nbl_tracker_jobs, 'get_job_output', side_effect=(
        'mock torrent filepath',
    ))

    assert nbl_tracker_jobs.torrent_filepath == 'mock torrent filepath'
    assert nbl_tracker_jobs.get_job_output.call_args_list == [
        call(nbl_tracker_jobs.create_torrent_job, slice=0),
    ]



from unittest.mock import Mock, call

import pytest

from upsies import errors
from upsies.jobs.set import SetJob


def config_mock(dct):
    def getitem(self, name):
        return dct[name]

    def setitem(self, name, value):
        dct[name] = value

    return Mock(
        paths=tuple(dct),
        __getitem__=getitem,
        __setitem__=setitem,
        keys=lambda: dct.keys(),
        values=lambda: dct.values(),
    )


@pytest.fixture
def make_SetJob(tmp_path):
    def make_SetJob(**kwargs):
        return SetJob(home_directory=tmp_path, cache_directory=tmp_path, **kwargs)
    return make_SetJob


def test_cache_id(make_SetJob):
    job = make_SetJob(config=config_mock({'foo': 'bar'}))
    assert job.cache_id is None


def test_output_is_hidden(make_SetJob):
    job = make_SetJob(config=config_mock({'foo': 'bar'}))
    assert job.hidden is True


def test_execute_finishes(make_SetJob):
    job = make_SetJob(config=config_mock({'foo': 'bar'}))
    assert not job.is_finished
    job.execute()
    assert job.is_finished


def test_no_arguments(mocker, make_SetJob):
    mocker.patch('upsies.jobs.set.SetJob._display_option')
    cfg = config_mock({'foo': '1', 'bar': '2', 'baz': '3'})
    job = make_SetJob(config=cfg)
    job.execute()
    assert job._display_option.call_args_list == [call('foo'), call('bar'), call('baz')]


@pytest.mark.parametrize('option', ('foo', 'bar', 'baz'))
def test_arguments_option(option, mocker, make_SetJob):
    mocker.patch('upsies.jobs.set.SetJob._display_option')
    values = {'foo': '1', 'bar': '2', 'baz': '3'}
    cfg = config_mock(values)
    job = make_SetJob(config=cfg, option=option)
    job.execute()
    assert cfg.reset.call_args_list == []
    assert cfg.write.call_args_list == []
    assert job._display_option.call_args_list == [call(option)]
    assert dict(cfg) == values

@pytest.mark.parametrize('option', ('foo', 'bar', 'baz'))
def test_arguments_option_and_value(option, mocker, make_SetJob):
    mocker.patch('upsies.jobs.set.SetJob._display_option')
    values = {'foo': '1', 'bar': '2', 'baz': '3'}
    cfg = config_mock(values)
    job = make_SetJob(config=cfg, option=option, value='hello')
    job.execute()
    assert cfg.reset.call_args_list == []
    assert cfg.write.call_args_list == [call(option)]
    assert job._display_option.call_args_list == [call(option)]
    assert dict(cfg) == {**values, option: 'hello'}

@pytest.mark.parametrize('option', ('foo', 'bar', 'baz'))
def test_arguments_option_and_reset(option, mocker, make_SetJob):
    mocker.patch('upsies.jobs.set.SetJob._display_option')
    values = {'foo': '1', 'bar': '2', 'baz': '3'}
    cfg = config_mock(values)
    job = make_SetJob(config=cfg, option=option, reset=True)
    job.execute()
    assert cfg.reset.call_args_list == [call(option)]
    assert cfg.write.call_args_list == [call(option)]
    assert job._display_option.call_args_list == [call(option)]
    assert dict(cfg) == values

def test_arguments_reset(mocker, make_SetJob):
    mocker.patch('upsies.jobs.set.SetJob._display_option')
    values = {'foo': '1', 'bar': '2', 'baz': '3'}
    cfg = config_mock(values)
    job = make_SetJob(config=cfg, reset=True)
    job.execute()
    assert cfg.reset.call_args_list == [call('foo'), call('bar'), call('baz')]
    assert cfg.write.call_args_list == [call('foo'), call('bar'), call('baz')]
    assert job._display_option.call_args_list == [call('foo'), call('bar'), call('baz')]
    assert dict(cfg) == values

def test_arguments_value(make_SetJob):
    with pytest.raises(RuntimeError, match=r'^Argument "value" needs argument "option"\.$'):
        make_SetJob(config=config_mock({'foo': 'bar'}), value='foo')

def test_arguments_value_and_reset(make_SetJob):
    with pytest.raises(RuntimeError, match=r'^Arguments "value" and "reset" are mutually exclusive\.$'):
        make_SetJob(config=config_mock({'foo': 'bar'}), value='foo', reset=True)

def test_arguments_option_and_value_and_reset(make_SetJob):
    with pytest.raises(RuntimeError, match=r'^Arguments "value" and "reset" are mutually exclusive\.$'):
        make_SetJob(config=config_mock({'foo': 'bar'}), value='foo', reset=True)


@pytest.mark.parametrize(
    argnames=('mode', 'kwargs'),
    argvalues=(
        ('_reset_mode', {'reset': True}),
        ('_set_mode', {'option': 'foo', 'value': 'bar'}),
        ('_display_mode', {}),
    ),
    ids=lambda v: str(v),
)
def test_ConfigError_is_handled(mode, kwargs, mocker, make_SetJob):
    mocker.patch('upsies.jobs.set.SetJob._display_option')
    mocker.patch(f'upsies.jobs.set.SetJob.{mode}', side_effect=errors.ConfigError('no!'))
    values = {'foo': '1', 'bar': '2', 'baz': '3'}
    cfg = config_mock(values)
    job = make_SetJob(config=cfg, **kwargs)
    job.execute()
    assert job.errors == (errors.ConfigError('no!'),)
    assert job.is_finished
    assert job.exit_code != 0
    assert cfg.reset.call_args_list == []
    assert cfg.write.call_args_list == []
    assert job._display_option.call_args_list == []
    assert dict(cfg) == values


@pytest.mark.parametrize(
    argnames=('name', 'value', 'exp_output'),
    argvalues=(
        ('foo', 'asdf', 'foo = asdf'),
        ('bar', 'a sd f', 'bar = a sd f'),
        ('baz', 123, 'baz = 123'),
        ('foo', ['a', 'b', 'c'], 'foo =\n  a\n  b\n  c'),
        ('foo', [1, 2, 3, 'd'], 'foo =\n  1\n  2\n  3\n  d'),
        ('foo', [], 'foo ='),
    ),
    ids=lambda v: str(v),
)
def test_display_option(name, value, exp_output, make_SetJob):
    job = make_SetJob(config=config_mock({name: value}))
    job.execute()
    assert job.output == (exp_output,)



import asyncio
import re
from unittest.mock import AsyncMock, Mock, PropertyMock, call

import pytest

from upsies import errors
from upsies.trackers import base


def make_MockTracker(**kwargs):
    class MockTracker(base.TrackerBase):
        name = 'asdf'
        label = 'AsdF'
        TrackerJobs = PropertyMock()
        TrackerConfig = PropertyMock()
        _login = AsyncMock()
        _logout = AsyncMock()
        get_announce_url = AsyncMock()
        upload = AsyncMock()

    return MockTracker(**kwargs)


def test_options():
    options = {'username': 'foo', 'password': 'bar'}
    tracker = make_MockTracker(options=options)
    assert tracker.options is options


def test_tasks():
    tracker = make_MockTracker()
    assert tracker._tasks == []
    assert tracker._tasks is tracker._tasks


@pytest.mark.parametrize('callback', (None, Mock()))
def test_attach_task(callback, mocker):
    tracker = make_MockTracker()
    run_task_mock = mocker.patch('upsies.utils.run_task')
    coro = Mock()
    if callback is None:
        return_value = tracker.attach_task(coro, callback=callback)
        assert run_task_mock.call_args_list == [call(coro, callback=callback)]
    else:
        return_value = tracker.attach_task(coro)
        assert run_task_mock.call_args_list == [call(coro, callback=None)]
    assert return_value is run_task_mock.return_value


@pytest.mark.asyncio
async def test_await_tasks(mocker):
    tracker = make_MockTracker()
    tasks = (
        asyncio.ensure_future(AsyncMock()()),
        asyncio.ensure_future(AsyncMock()()),
        asyncio.ensure_future(AsyncMock()()),
    )
    tracker._tasks.extend(tasks)

    for task in tasks:
        assert not task.done()

    await tracker.await_tasks()

    for task in tasks:
        assert task.done()


@pytest.mark.asyncio
async def test_login_when_already_logged_in(mocker):
    tracker = make_MockTracker()
    mocker.patch.object(tracker, '_login')
    mocker.patch.object(tracker, '_is_logged_in', True, create=True)

    assert tracker.is_logged_in

    # Login multiple times concurrently
    await asyncio.gather(
        tracker.login(),
        tracker.login(),
        tracker.login(),
    )

    assert tracker.is_logged_in
    assert tracker._login.call_args_list == []

@pytest.mark.asyncio
async def test_login_succeeds(mocker):
    tracker = make_MockTracker()
    mocker.patch.object(tracker, '_login')

    assert not tracker.is_logged_in

    # Login multiple times concurrently
    await asyncio.gather(
        tracker.login(),
        tracker.login(),
        tracker.login(),
    )

    assert tracker.is_logged_in
    assert tracker._login.call_args_list == [call()]

@pytest.mark.asyncio
async def test_login_fails(mocker):
    tracker = make_MockTracker()
    exception = errors.RequestError('nope')
    mocker.patch.object(tracker, '_login', side_effect=exception)

    assert not tracker.is_logged_in

    # Login multiple times concurrently
    with pytest.raises(type(exception), match=rf'^{re.escape(str(exception))}$'):
        await asyncio.gather(
            tracker.login(),
            tracker.login(),
            tracker.login(),
        )

    assert not tracker.is_logged_in
    assert tracker._login.call_args_list == [call(), call(), call()]

@pytest.mark.asyncio
async def test_logout_when_already_logged_out(mocker):
    tracker = make_MockTracker()
    mocker.patch.object(tracker, '_logout')
    mocker.patch.object(tracker, '_is_logged_in', False, create=True)

    assert not tracker.is_logged_in

    # Logout multiple times concurrently
    await asyncio.gather(
        tracker.logout(),
        tracker.logout(),
        tracker.logout(),
    )

    assert not tracker.is_logged_in
    assert tracker._logout.call_args_list == []

async def test_logout_succeeds(mocker):
    tracker = make_MockTracker()
    mocker.patch.object(tracker, '_logout')
    mocker.patch.object(tracker, '_is_logged_in', True, create=True)

    assert tracker.is_logged_in

    # Logout multiple times concurrently
    await asyncio.gather(
        tracker.logout(),
        tracker.logout(),
        tracker.logout(),
    )

    assert not tracker.is_logged_in
    assert tracker._logout.call_args_list == [call()]

@pytest.mark.asyncio
async def test_logout_fails(mocker):
    tracker = make_MockTracker()
    exception = errors.RequestError('nope')
    mocker.patch.object(tracker, '_logout', side_effect=exception)
    mocker.patch.object(tracker, '_is_logged_in', True, create=True)

    assert tracker.is_logged_in

    # Logout multiple times concurrently
    with pytest.raises(type(exception), match=rf'^{re.escape(str(exception))}$'):
        await asyncio.gather(
            tracker.logout(),
            tracker.logout(),
            tracker.logout(),
        )

    assert not tracker.is_logged_in
    assert tracker._logout.call_args_list == [call()]


def test_calculate_piece_size():
    tracker = make_MockTracker()
    assert type(tracker).calculate_piece_size is None


def test_calculate_piece_size_min_max():
    tracker = make_MockTracker()
    assert type(tracker).calculate_piece_size_min_max is None


def test_signals():
    tracker = make_MockTracker()
    cb = Mock()
    tracker.signal.register('warning', cb.warn)
    tracker.signal.register('error', cb.error)
    tracker.signal.register('exception', cb.exception)
    tracker.warn('foo')
    assert cb.mock_calls == [call.warn('foo')]
    tracker.error('bar')
    assert cb.mock_calls == [call.warn('foo'), call.error('bar')]
    tracker.exception('baz')
    assert cb.mock_calls == [call.warn('foo'), call.error('bar'), call.exception('baz')]


def test_generate_setup_howto_makes_expected_calls(mocker):
    Howto_mock = mocker.patch('upsies.trackers.base.tracker.Howto')
    evaluate_fstring_mock = mocker.patch('upsies.utils.string.evaluate_fstring')
    tracker = make_MockTracker()
    return_value = type(tracker).generate_setup_howto()
    assert return_value is evaluate_fstring_mock.return_value
    assert Howto_mock.call_args_list == [call(tracker_cls=type(tracker))]

def test_generate_setup_howto_populates_namespace_expectedly(mocker):
    tracker = make_MockTracker()
    type(tracker).setup_howto_template = '{isinstance(howto, Howto)}'
    assert type(tracker).generate_setup_howto() == 'True'



import pytest


@pytest.fixture
def assert_config_choice():
    def assert_config_choice(choice, value, options, description=None):
        assert choice == value
        assert choice.options == tuple(options)

        cls = type(choice)
        for option in options:
            assert cls(option) == option
            assert cls(option).options == tuple(options)

        options_string = ', '.join(options)
        with pytest.raises(ValueError, match=rf'^Not one of {options_string}: foo$'):
            cls('foo')

        if description is not None:
            assert choice.description == description

    return assert_config_choice


@pytest.fixture
def assert_config_number():
    def assert_config_number(number, value, min, max, description=None):
        assert number == value
        assert number.min == min
        assert number.max == max

        cls = type(number)
        for i in range(min, max + 1):
            assert cls(i) == i
            assert cls(i).min == min
            assert cls(i).max == max

        with pytest.raises(ValueError, match=rf'^Minimum is {min}$'):
            cls(min - 1)

        with pytest.raises(ValueError, match=rf'^Maximum is {max}$'):
            cls(max + 1)

        if description is not None:
            assert number.description == description

    return assert_config_number
